import string
import random
import phenopackets
from .variant import Variant

ACCEPTABLE_GENOMES = {"GRCh37", "GRCh38", "hg19", "hg38"}


class StructuralVariant(Variant):
    """
    This encapsulates variant about a structural variant
    For instance, we may see things like this
    chr7.hg19:g.(35674000_37280000)_(46_111_000_46_598_000)del (fuzzy boundaries)
    chr7.hg19:g(38521704_45810267)del (precise boundaries)
    rsa7p14.1(kit P179)x1
    46,XY.ish del(7)(p14.1)(RP11-816F16-)
    46,XX.ish del(7)(p14.1p14.1)(GLI3-)
    46,XY.ish del(7)(p14.1)(GLI3-)[56]/7p14.1(GLI3x2)[44] (various ICSN or bespoke)
    Our strategy is not to model these variants precisely. Instead, for genotype-phenotype
    analysis, it may be enough to know that these are structural variants and thus
    likely to be complete loss of function.
    So we record something about the variant and add it by hand to the Individual object
    We want to be able to create a GA4GH VariationDescriptor object with the following fields
    - id - required, autogenerate if user provides no id
    - variant (VRS) -- leave this field empty
    - label -- the original contents of the cell, e.g., 46,XY.ish del(7)(p14.1)(RP11-816F16-)
    - gene context -- the gene that is disrupted by the structural variant
    - expressions, empty
    - vcf_record, empty
    - structural type: Ontology Class
    - allelic state: het/hom/emi etc.

    :param cell_contents: the string from the original table that we want to map as a structural variant
    :type cell_contents: str
    :param gene_symbol: the gene affected by the structural variant, e.g., GLI3
    :type gene_symbol: str
    :param gene_id: the identifier (using HGNC) of the gene, e.g., GLI3 is HGNC:4319
    :type gene_id: str
    :param sequence_ontology_id: An identifier from the Sequence Ontology
    :type sequence_ontology_id: str
    :param sequence_ontology_label: the SO label corresponding to the ID
    :type sequence_ontology_label: str
    :param variant_id: an identifier for the variant, optional
    :type variant_id: str
    """

    def __init__(self, cell_contents,
                 gene_symbol,
                 gene_id,
                 sequence_ontology_id,
                 sequence_ontology_label,
                 variant_id: None):
        super().__init__()
        if variant_id is None:
            self._variant_id = "var_" + "".join(random.choices(string.ascii_letters, k=25))
        else:
            self._variant_id = variant_id
        self._label = cell_contents.strip()
        if gene_symbol is None:
            raise ValueError(f"Need to pass a valid gene symbol!")
        self._gene_symbol = gene_symbol
        if gene_id is None:
            raise ValueError(f"Need to pass a valid HGNC gene id!")
        self._hgnc_id = gene_id
        self._so_id = sequence_ontology_id
        self._so_label = sequence_ontology_label
        self._genotype = None

    def to_ga4gh_variant_interpretation(self, acmg=None):
        """
        Transform this Variant object into a "variantInterpretation" message of the GA4GH Phenopacket schema
        """
        vdescriptor = phenopackets.VariationDescriptor()
        vdescriptor.id = self._variant_id
        vdescriptor.gene_context.value_id = self._hgnc_id
        vdescriptor.gene_context.symbol = self._gene_symbol
        vdescriptor.label = self._label
        vdescriptor.molecule_context = phenopackets.MoleculeContext.genomic
        if self._genotype is not None:
            if self._genotype == 'heterozygous':
                vdescriptor.allelic_state.id = "GENO:0000135"
                vdescriptor.allelic_state.label = "heterozygous"
            elif self._genotype == 'homozygous':
                vdescriptor.allelic_state.id = "GENO:0000136"
                vdescriptor.allelic_state.label = "homozygous"
            elif self._genotype == 'hemizygous':
                vdescriptor.allelic_state.id = "GENO:0000134"
                vdescriptor.allelic_state.label = "hemizygous"
            else:
                print(f"Did not recognize genotype {self._genotype}")
        vdescriptor.structural_type.id = self._so_id
        vdescriptor.structural_type.label = self._so_label
        vinterpretation = phenopackets.VariantInterpretation()
        if acmg is not None:
            if acmg.lower() == 'benign':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.BENIGN
            elif acmg.lower == 'likely benign' or acmg.lower() == 'likely_benign':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.LIKELY_BENIGN
            elif acmg.lower == 'uncertain significance' or acmg.lower() == 'uncertain_significance':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.UNCERTAIN_SIGNIFICANCE
            elif acmg.lower == 'likely pathogenic' or acmg.lower() == 'likely_pathogenic':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.LIKELY_PATHOGENIC
            elif acmg.lower == 'pathogenic' or acmg.lower() == 'pathogenic':
                vinterpretation.acmgPathogenicityClassification = phenopackets.AcmgPathogenicityClassification.PATHOGENIC
            else:
                print(f"Warning- did not recognize ACMG category {acmg}")
        vinterpretation.variation_descriptor.CopyFrom(vdescriptor)
        return vinterpretation

    # provide static constructors for most important structural variation types

    @staticmethod
    def chromosomal_deletion(cell_contents,
                             gene_symbol,
                             gene_id,
                             variant_id=None):
        """
        create a StructuralVariant object for a chromosomal deletion

        :param cell_contents: the string from the original table that we want to map as a structural variant
        :type cell_contents: str
        :param gene_symbol: the gene affected by the structural variant, e.g., GLI3
        :type gene_symbol: str
        :param gene_id: the identifier (using HGNC) of the gene, e.g., GLI3 is HGNC:4319
        :type gene_id: str
        :param variant_id: an identifier for the variant
        :type variant_id: str, optional
        """
        return StructuralVariant(cell_contents=cell_contents,
                                 gene_symbol=gene_symbol,
                                 gene_id=gene_id,
                                 sequence_ontology_id="SO:1000029",
                                 sequence_ontology_label="chromosomal_deletion",
                                 variant_id=variant_id)

    @staticmethod
    def chromosomal_duplication(cell_contents,
                                gene_symbol,
                                gene_id,
                                variant_id=None):
        """
        create a StructuralVariant object for a chromosomal duplication

        :param cell_contents: the string from the original table that we want to map as a structural variant
        :type cell_contents: str
        :param gene_symbol: the gene affected by the structural variant, e.g., GLI3
        :type gene_symbol: str
        :param gene_id: the identifier (using HGNC) of the gene, e.g., GLI3 is HGNC:4319
        :type gene_id: str
        :param variant_id: an identifier for the variant
        :type variant_id: str, optional
        """
        return StructuralVariant(cell_contents=cell_contents,
                                 gene_symbol=gene_symbol,
                                 gene_id=gene_id,
                                 sequence_ontology_id="SO:1000037",
                                 sequence_ontology_label="chromosomal_duplication",
                                 variant_id=variant_id)

    @staticmethod
    def chromosomal_inversion(cell_contents,
                              gene_symbol,
                              gene_id,
                              genotype,
                              variant_id=None):
        """
        create a StructuralVariant object for a chromosomal inversion

        :param cell_contents: the string from the original table that we want to map as a structural variant
        :type cell_contents: str
        :param gene_symbol: the gene affected by the structural variant, e.g., GLI3
        :type gene_symbol: str
        :param gene_id: the identifier (using HGNC) of the gene, e.g., GLI3 is HGNC:4319
        :type gene_id: str
        :param genotype: Genotype (heterozygous, homozygous, hemizygous) of this variant call
        :type genotype: str
        :param variant_id: an identifier for the variant
        :type variant_id: str, optional
        """
        return StructuralVariant(cell_contents=cell_contents,
                                 gene_symbol=gene_symbol,
                                 gene_id=gene_id,
                                 sequence_ontology_id="SO:1000030",
                                 sequence_ontology_label="chromosomal_inversion",
                                 variant_id=variant_id)



from .cohort_validator import CohortValidator
from .content_validator import ContentValidator
from .ontology_qc import OntologyQC
from .phenopacket_validator import PhenopacketValidator
from .validation_result import ValidationResult, ValidationResultBuilder

__all__ = [
    "CohortValidator",
    "ContentValidator",
    "OntologyQC",
    "PhenopacketValidator",
    "ValidationResult",
    "ValidationResultBuilder"
]



import unittest
import os
from src.pyphetools.visualization import PhenopacketTable

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')


class TestPhenopacketTable(unittest.TestCase):

    def test_age_1(self):
        iso8601 = "P7Y"
        expected_days = int(365.25 * 7)
        days = PhenopacketTable.iso_to_days(iso_age=iso8601)
        self.assertEqual(expected_days, days)

    def test_age_2(self):
        iso8601 = "P7Y2M"
        expected_days = int(365.25 * 7) + int(30.437*2)
        days = PhenopacketTable.iso_to_days(iso_age=iso8601)
        self.assertEqual(expected_days, days)

    def test_age_3(self):
        iso8601 = "P7Y2M6D"
        expected_days = int(365.25 * 7) + int(30.437*2) + 6
        days = PhenopacketTable.iso_to_days(iso_age=iso8601)
        self.assertEqual(expected_days, days)



from .simple_patient import SimplePatient
import pandas as pd
from .hpo_category import HpoCategorySet
from hpotk.model import TermId
from collections import defaultdict

ALL_ROOT = TermId.from_curie("HP:0000001")
PHENOTYPIC_ABNORMALITY_ROOT = TermId.from_curie("HP:0000118")



class DetailedSupplTable:
    """
    This class intends to facilitate the creation of a detailed supplemental tabular file with the phenotypic findings for
    an entire cohort. Each attribute (id, sex, age, HPO terms) are shown on rows, and the individuals who make up the cohort
    are shown as columns.
    """

    def __init__(self, patient_d, hp_ontology) -> None:
        """
        :param patient_d: dictionary of patients to display
        :type patient_d: map with key string and value SimplePatient
        """
        if not isinstance(patient_d, dict):
            raise ValueError(f"patient_d argument must be dictionary but was {type(patient_d)}")
        my_simple_patient_d = defaultdict(SimplePatient)
        for k, v in patient_d.items():
            if str(type(v)) == "<class 'phenopackets.schema.v2.phenopackets_pb2.Phenopacket'>":
                my_simple_patient_d[k] = SimplePatient(ga4gh_phenopacket=v)
            else:
                raise ValueError(f"patient_d values must be GA4GH Phenopackets but was {type(v)}")
        self._patient_d = patient_d  # GA4GH phenopackets 
        self._simple_patient_d = my_simple_patient_d # SimplePatients
        self._hp_ontology = hp_ontology
        self._total_counts = defaultdict(int)
        self._total_counts_propagated = defaultdict(int)
        sex_d = defaultdict(int)
        var_d = defaultdict(int)
        for pat_id, pat in self._simple_patient_d.items():
            sex_d[pat.get_sex()] += 1
            hpo_terms = pat.get_observed_hpo_d()
            anc_set = set() # graph with ancestors induced by all terms of the patient
            for hp_id in hpo_terms.keys():
                # key is a string such as HP:0001234, value is an HpTerm object
                # we need to convert it to an object from hpo-toolkit because get_ancestors returns HpTerm objects
                hp_termid = TermId.from_curie(hp_id)
                ancs = self._hp_ontology.graph.get_ancestors(hp_termid)
                anc_set.add(hp_termid)
                anc_set.update(ancs)
                self._total_counts[hp_id] += 1
            for hp_id in anc_set:
                if hp_id == ALL_ROOT or hp_id == PHENOTYPIC_ABNORMALITY_ROOT:
                    continue
                else:
                    self._total_counts_propagated[hp_id.value] += 1
            variants = pat.get_variant_list()
            for var in variants:
                var_d[var] += 1

            # TODO figure out what to do with biallelic
        self._hpo_category_set = HpoCategorySet(ontology=hp_ontology)


    def _get_table(self, counts_d):
        """
        Get counts of terms without annotation propagation or thresholding
        """
        rows = []
        N = len(self._patient_d)
        for hpid, total_count in counts_d.items():
            total_per = 100*total_count/N
            total_s = f"{total_count}/{N} ({total_per:.1f}%)"
            hpterm = self._ontology.get_term(hpid)
            cat = self.get_category(termid=hpid)
            focus_count = self._focus_counts.get(hpid, 0)
            other_count = self._non_focus_counts.get(hpid, 0)
            d = {'category': cat, 'term': hpterm.name, 'HP:id': hpid, 'focus' : focus_count, 'other': other_count, 'total': total_s, 'total_count': total_count}
            rows.append(d)
        df = pd.DataFrame(rows)
        df.set_index('category', inplace=True)
        return df.sort_values(['category', 'total_count'], ascending=[True, False])


    
    def get_table_direct_annotations(self):
        """
        Get counts of terms without annotation propagation or thresholding
        """
        return self._get_table(self._total_counts)
    

    def get_table_with_propagated_counts(self):
        """
        Get counts of terms without annotation propagation or thresholding
        """
        return self._get_table(self._total_counts_propagated)
    
    @staticmethod
    def _get_counts(hpo_term_id:str, counts_d:dict):
        M = len(counts_d)
        N = 0
        for k, v in counts_d.items():
            if v.contains_observed_term_id(hpo_term_id):
                N += 1
        return N, M
    
    @staticmethod
    def _get_counts_dict(pat_list):
        counts_d = defaultdict(int)
        for pat in pat_list:
            observed_hpo = pat.get_observed_hpo_d()
            for hpo_term_id in observed_hpo.keys():
                counts_d[hpo_term_id] += 1
        return counts_d   
    

    def get_html_table_by_pmid(self, min_count=0):
        """
        :param min_count: minimum count to be displayed ion the table (by default, all terms are displayed)
        :type min_count: int
        """
        by_pmid_d = defaultdict(list)
        all_observed_hpo = defaultdict(int) # HPO ids observed in our cohort
        hpo_id_to_display_d = defaultdict()
        for pat in self._simple_patient_d.values():
            observed_hpo_d = pat.get_observed_hpo_d()
            for hpo_term in observed_hpo_d.values():
                hpo_term_id = hpo_term.id
                all_observed_hpo[hpo_term_id] += 1
                hpo_id_to_display_d[hpo_term_id] = hpo_term
            if pat.has_pmid():
                by_pmid_d[pat.get_pmid()].append(pat) 
            else:
                by_pmid_d["n/a"].append(pat) 
        sorted_pmids = sorted(by_pmid_d.items(), key=lambda x:len(x[1]), reverse=True) # sort by values
        sorted_pmids = [ x[0] for x  in sorted_pmids]
        sorted_hpos = sorted(all_observed_hpo.items(), key=lambda x:x[1], reverse=True)
        sorted_hpos = [ x[0] for x  in sorted_hpos if x[1] > min_count]
        table_items = []
        table_items.append('<table style="border: 2px solid black;">\n')
        middle = []
        for pmid in sorted_pmids:
            middle.append(f"<th>{pmid}</th>")
        middle_txt = "".join(middle)
        header = f"<tr><th>HPO term</th>{middle_txt}</tr>"
        table_items.append(header)
        for hpo_term_id in sorted_hpos:
            hpo_term = hpo_id_to_display_d.get(hpo_term_id)
            line_items = []
            line_items.append(f"<tr><td>{hpo_term}</td>")
            for pmid in sorted_pmids:
                simple_pat_list = by_pmid_d.get(pmid)
                counts_d = DetailedSupplTable._get_counts_dict(simple_pat_list)
                M = len(simple_pat_list)
                N = counts_d.get(hpo_term_id, 0)
                cell_contents = f"{N}/{M} ({100*N/M:.1f}%)"
                line_items.append(f"<td>{cell_contents}</td>")
            line_items.append("</tr>\n")
            table_items.append("".join(line_items))
        table_items.append('</table>\n') # close table content
        return "\n".join(table_items)
        

            
        
        
        



import unittest
import os
import phenopackets as PPkt
from src.pyphetools.visualization import SimpleVariant
from src.pyphetools.creation import StructuralVariant

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')

class TestHpoParser(unittest.TestCase):


    def test_structural_variant_display(self):
        eri1_id = "HGNC:23994"
        eri1_sumbol = "ERI1"
        sv = StructuralVariant.chromosomal_deletion(cell_contents="g.8783887_9068578del", gene_id=eri1_id, gene_symbol=eri1_sumbol)
        variant_interpretation = sv.to_ga4gh_variant_interpretation()
        genomic_interpretation = PPkt.GenomicInterpretation()
        genomic_interpretation.subject_or_biosample_id = "some.id"
        # by assumption, variants passed to this package are all causative
        genomic_interpretation.interpretation_status = PPkt.GenomicInterpretation.InterpretationStatus.CAUSATIVE
        genomic_interpretation.variant_interpretation.CopyFrom(variant_interpretation)
        simpleVar = SimpleVariant(genomic_interpretation)
        self.assertEqual("g.8783887_9068578del: chromosomal_deletion (SO:1000029)", simpleVar.get_display())


import unittest
import os
from src.pyphetools.creation import HpoParser, OptionColumnMapper

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')


class TestAssumeExcluded(unittest.TestCase):
    """
    This test class checks the function of the assumeExcluded argument of the OptionColumnMapper.
    """

    @classmethod
    def setUpClass(cls) -> None:
        parser = HpoParser(hpo_json_file=HP_JSON_FILENAME)
        cls.hpo_cr = parser.get_hpo_concept_recognizer()
        cls.severity_d = {'mild': 'Intellectual disability, mild',
                        'moderate': 'Intellectual disability, moderate',
                        'severe': 'Intellectual disability, severe'
                        }


    def test_without_assume_exclusion(self):
        comorbidities_d = {'Insulindependentdiabetes-onsetatage19': 'Type II diabetes mellitus',
                            'COPD': 'Chronic pulmonary obstruction',
                            'Diastolicdysfunction': 'Left ventricular diastolic dysfunction',
                            'Lymphocytosis': 'Lymphocytosis',
                            'Single kidney': 'Unilateral renal agenesis'}
        comorbiditiesMapper = OptionColumnMapper(column_name="placeholder",
                                                concept_recognizer=self.hpo_cr,
                                                option_d=comorbidities_d)
        results = comorbiditiesMapper.map_cell('COPD')
        self.assertEqual(1, len(results))
        hpo_term = results[0]
        self.assertEqual('Chronic pulmonary obstruction', hpo_term.label)
        self.assertEqual('HP:0006510', hpo_term.id)

    def test_with_assume_exclusion(self):
        excluded_d = {'Insulindependentdiabetes-onsetatage19': 'Type II diabetes mellitus',
                            'COPD': 'Chronic pulmonary obstruction',
                            'Diastolicdysfunction': 'Left ventricular diastolic dysfunction',
                            'Lymphocytosis': 'Lymphocytosis',
                            'Single kidney': 'Unilateral renal agenesis'}
        comorbiditiesMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d={}, excluded_d=excluded_d)
        results = comorbiditiesMapper.map_cell('COPD')
        self.assertEqual(1, len(results))
        hpo0 = results[0]
        self.assertEqual('Chronic pulmonary obstruction', hpo0.label)
        self.assertEqual('HP:0006510', hpo0.id)
        self.assertFalse(hpo0.observed)
        results = comorbiditiesMapper.map_cell('Diastolicdysfunction')
        hpo1 = results[0]
        self.assertEqual("Left ventricular diastolic dysfunction", hpo1.label)
        self.assertEqual("HP:0025168", hpo1.id)
        self.assertFalse(hpo1.observed)
        results = comorbiditiesMapper.map_cell('Lymphocytosis')
        hpo2 = results[0]
        self.assertEqual("Lymphocytosis", hpo2.label)
        self.assertFalse(hpo2.observed)
        results = comorbiditiesMapper.map_cell('Insulindependentdiabetes-onsetatage19')
        hpo3 = results[0]
        self.assertEqual("Type II diabetes mellitus", hpo3.label)
        self.assertFalse(hpo3.observed)
        results = comorbiditiesMapper.map_cell('Single kidney')
        hpo4 = results[0]
        self.assertEqual("Unilateral renal agenesis", hpo4.label)
        self.assertFalse(hpo4.observed)



import math
from .constants import Constants

AVERAGE_DAYS_IN_MONTH = 30.437

class AgeIsoFormater:
    """
    A helper class that formats year, month, week, days into ISO 8601 strings

    :param y: years
    :type y: Union(int,str), optional
    :param m: months
    :type m: Union(int,str), optional
    :param w: weeks
    :type w: Union(int,str), optional
    :param d: days
    :type d: Union(int,str), optional


    """

    def __init__(self, y=None, m=None, w=None, d=None):
        if y is None:
            years = 0
        else:
            years = int(y)
        if m is None:
            months = 0
        else:
            months = int(m)
        if w is not None:
            extradays = 7 * int(w)
        else:
            extradays = 0
        if d is None:
            days = 0 + extradays
        else:
            days = int(d) + extradays
        if days > AVERAGE_DAYS_IN_MONTH:
            extra_months = math.floor(days / AVERAGE_DAYS_IN_MONTH)
            months = months + int(extra_months)
            days = days % int(AVERAGE_DAYS_IN_MONTH)  # modulo arithmetic, get remaining days after months are subtracted
        if months > 11:
            extra_years = months // 12
            months = months % 12
            years = years + extra_years
        self._years = years
        self._months = months
        self._days = days

    def to_iso8601(self):
        components = ["P"]
        if self._years > 0:
            components.append(f"{self._years}Y")
        if self._months > 0:
            components.append(f"{self._months}M")
        if self._days > 0:
            components.append(f"{self._days}D")
        if len(components) == 1:
            return "P0D" # newborn
        else:
            return "".join(components)

    @staticmethod
    def to_string(y=None, m=None, w=None, d=None):
        """
        :param y: years
        :type y: Union(int,str), optional
        :param m: months
        :type m: Union(int,str), optional
        :param w: weeks
        :type w: Union(int,str), optional
        :param d: days
        :type d: Union(int,str), optional
        :returns: ISO8601 string representing age
        :rtype: str
        """
        formater = AgeIsoFormater(y=y, m=m, w=w, d=d)
        return formater.to_iso8601()


    @staticmethod
    def from_numerical_month(month):
        """
        decode entries such as 18 or 0.7 (number of months)
        """
        month = str(month)
        if month.isdigit():
            full_months = int(month)
            days = 0
        elif month.replace('.', '', 1).isdigit() and month.count('.') < 2:
            # a float such as 0.9 (months)
            months = float(month)
            avg_num_days_in_month = 30.437
            floor_months = math.floor(months)
            if floor_months == 0.0:
                days = int(months * avg_num_days_in_month)
                full_months = 0
            else:
                remainder = months - floor_months
                full_months = int(months - remainder)
                days = int(remainder * avg_num_days_in_month)
        else:
            return Constants.NOT_PROVIDED
        formater = AgeIsoFormater(m=full_months, d=days)
        return formater.to_iso8601()






import unittest
from src.pyphetools.creation import HgvsVariant

# The following dictionary has the same structure as a subelement of the response from Variant Validator
VCF_DICTIONARY =  {'alt': 'C', 'chr': '16', 'pos': '1756403', 'ref': 'CG'}
GENOME_ASSEMBLY = 'hg38'
transcript = 'NM_015133.4'
symbol = 'MAPK8IP3'
hgvs = 'NM_015133.4:c.111C>G'
hgnc_id = 'HGNC:6884'
g_hgvs = 'NC_000016.10:g.1706450C>G'

class TestVariantValidator(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls._variant = HgvsVariant(assembly=GENOME_ASSEMBLY, vcf_d=VCF_DICTIONARY, symbol=symbol,
                                    hgnc=hgnc_id, hgvs=hgvs, transcript=transcript, g_hgvs=g_hgvs)

    def test_obect_created(self):
        self.assertIsNotNone(self._variant)

    def test_assembly(self):
        var = self._variant
        self.assertEqual(GENOME_ASSEMBLY, var.assembly)

    def test_chr(self):
        var = self._variant
        self.assertEqual('16', var.chr)

    def test_position(self):
        self.assertEqual(1756403, self._variant.position)

    def test_ref(self):
        var = self._variant
        self.assertEqual('CG', var.ref)

    def test_alt(self):
        var = self._variant
        self.assertEqual('C', var.alt)

    def test_to_ga4gh(self):
        variant_descriptor = self._variant.to_ga4gh()
        self.assertIsNotNone(variant_descriptor)




import json
import os
from typing import List, Optional

import phenopackets
from google.protobuf.json_format import Parse

from .phenopacket_validator import PhenopacketValidator
from .validation_result import ValidationResult, ValidationResultBuilder
from ..creation.allelic_requirement import AllelicRequirement
from ..creation.individual import Individual



class ContentValidator(PhenopacketValidator):
    """
    Validate a list of phenopackets as to whether they have a minunum number of phenotypic features and alleles

    The following example shows how to use this class to assess whether each phenopacket in the directory called "phenopackets" contains at least one variant and at least three HPO terms.

        from pyphetools.visualization import PhenopacketIngestor
        from pyphetools.validation import ContentValidator
        ingestor = PhenopacketIngestor(indir="phenopackets")
        ppkt_d = ingestor.get_phenopacket_dictionary()
        ppkt_list = list(ppkt_d.values())
        validator = ContentValidator(min_var=1, min_hpo=3)
        errors = validator.validate_phenopacket_list(ppkt_list)
        print(f"{len(errors)} errors were identified")

    Note that this class does not test for all errors. Use phenopacket-tools to check for redundant or conflicting
    annotations.

    :param min_hpo: minimum number of phenotypic features (HP terms) for this phenopacket to be considered valid
    :type min_hpo: int
    :param allelic_requirement: used to check number of alleles and variants
    :type allelic_requirement: AllelicRequirement
    """
    def __init__(self, min_hpo: int, allelic_requirement: AllelicRequirement = None) -> None:
        super().__init__()
        self._min_hpo = min_hpo
        self._allelic_requirement = allelic_requirement


    def validate_individual(self, individual:Individual) -> List[ValidationResult]:
        """
        check a single Individual as to whether there are sufficient HPO terms and alleles/variants
        :returns: a potential empty list of validations
        :rtype: List[ValidationResult]
        """
        n_pf = len(individual.hpo_terms)
        n_var = 0
        n_alleles = 0
        pp_id = individual.get_phenopacket_id()
        for variant_interpretation in individual.interpretation_list:
            n_var += 1
            if variant_interpretation.variation_descriptor is not None:
                vdesc =  variant_interpretation.variation_descriptor
                if vdesc.allelic_state is not None:
                    gtype = vdesc.allelic_state
                    if gtype.label == "heterozygous": # "GENO:0000135"
                        n_alleles += 1
                    elif gtype.label == "homozygous": # "GENO:0000136"
                        n_alleles += 2
                    elif gtype.label == "hemizygous": # "GENO:0000134"
                        n_alleles += 1
        return self._validate(pp_id=pp_id, n_hpo=n_pf, n_var=n_var, n_alleles=n_alleles)



    def validate_phenopacket(self, phenopacket) -> List[ValidationResult]:
        """
        check a single phenopacket as to whether there are sufficient HPO terms and alleles/variants
        :returns: a potential empty list of validations
        :rtype: List[ValidationResult]
        """
        if isinstance(phenopacket, str):
            # the user passed a file
            if not os.path.isfile(phenopacket):
                raise FileNotFoundError(f"Could not find phenopacket file at '{phenopacket}'")
            with open(phenopacket) as f:
                data = f.read()
                jsondata = json.loads(data)
                phpacket = Parse(json.dumps(jsondata), phenopackets.Phenopacket())
        elif isinstance(phenopacket, phenopackets.Phenopacket):
            phpacket = phenopacket
        else:
            raise ValueError(f"phenopacket argument must be file path or GA4GH Phenopacket \
                object but was {type(phenopacket)}")
        pp_id = phpacket.id
        n_pf = len(phpacket.phenotypic_features)
        if phpacket.interpretations is None:
            n_var = 0
            n_alleles = 0
        else:
            n_var = 0
            n_alleles = 0
            for interpretation in phpacket.interpretations:
                if interpretation.diagnosis is not None:
                    dx = interpretation.diagnosis
                    for genomic_interpretation in dx.genomic_interpretations:
                        n_var += 1
                        vint = genomic_interpretation.variant_interpretation
                        if vint.variation_descriptor is not None:
                            vdesc =   vint.variation_descriptor
                            if vdesc.allelic_state is not None:
                                gtype = vdesc.allelic_state
                                if gtype.label == "heterozygous": # "GENO:0000135"
                                    n_alleles += 1
                                elif gtype.label == "homozygous": # "GENO:0000136"
                                    n_alleles += 2
                                elif gtype.label == "hemizygous": # "GENO:0000134"
                                    n_alleles += 1
        return self._validate(pp_id=pp_id, n_hpo=n_pf, n_var=n_var, n_alleles=n_alleles)



    def _validate(self, pp_id:str, n_hpo:int, n_var:int=None, n_alleles:int=None):
        """
        private method called by validate_individual or validate_phenopacket.
        :param pp_id: phenopacket identifier
        :type pp_id: str
        :param n_hpo: Number of HPO terms
        :type n_hpo: int
        :param n_var: Number of variants found
        :type n_var: Optional[int]
        :param n_alleles: Number of alleles found
        :type n_alleles: Optional[int]
        """
        validation_results = []
        if n_hpo < self._min_hpo:
            validation_results.append(ValidationResultBuilder(phenopacket_id=pp_id).insufficient_hpos(min_hpo=self._min_hpo, n_hpo=n_hpo).build())
        if self._allelic_requirement is None:
            return validation_results
        if self._allelic_requirement == AllelicRequirement.MONO_ALLELIC:
            if n_var != 1:
                val_result = ValidationResultBuilder(phenopacket_id=pp_id).incorrect_variant_count(self._allelic_requirement, n_var).build()
                validation_results.append(val_result)
            if n_alleles != 1:
                val_result = ValidationResultBuilder(phenopacket_id=pp_id).incorrect_allele_count(self._allelic_requirement, n_alleles).build()
                validation_results.append(val_result)
        elif self._allelic_requirement == AllelicRequirement.BI_ALLELIC:
            if n_var < 1 or n_var > 2:
                msg = f"Expected one or two variant for biallelic but got {n_var} variants"
                val_result = ValidationResultBuilder(phenopacket_id=pp_id).incorrect_variant_count(self._allelic_requirement, n_var).build()
                validation_results.append(val_result)
            if n_alleles != 2:
                val_result = ValidationResultBuilder(phenopacket_id=pp_id).incorrect_allele_count(self._allelic_requirement, n_alleles).build()
                validation_results.append(val_result)
        return validation_results






import os
import unittest

from src.pyphetools.creation import HpoParser

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')


class TestOptionMapper(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        parser = HpoParser(hpo_json_file=HP_JSON_FILENAME)
        cls.hpo_cr = parser.get_hpo_concept_recognizer()

    def test_mixed(self):
        cell_contents = "Long and thick eyebrows, upper slanted palpebral fissures, anteverted nares, short philtrum"
        other_d = {'upper slanted palpebral fissures': 'Upslanted palpebral fissure'}
        results = self.hpo_cr.parse_cell(cell_contents=cell_contents, custom_d=other_d)
        # sort results in order of HP ids for testing
        # We expect
        # Thick eyebrow (HP:0000574)
        # Upslanted palpebral fissure (HP:0000582)
        # Anteverted nares (HP:0000463)
        # Short philtrum (HP:0000322)
        results = sorted(results, key=lambda x: x.id)
        self.assertEqual(4, len(results))
        self.assertEqual("HP:0000322", results[0].id)
        self.assertEqual("Short philtrum", results[0].label)
        self.assertTrue(results[0].observed)
        self.assertTrue(results[0].measured)
        self.assertEqual("HP:0000463", results[1].id)
        self.assertEqual("Anteverted nares", results[1].label)
        self.assertTrue(results[1].observed)
        self.assertTrue(results[1].measured)
        self.assertEqual("HP:0000574", results[2].id)
        self.assertEqual("Thick eyebrow", results[2].label)
        self.assertTrue(results[2].observed)
        self.assertTrue(results[2].measured)
        self.assertEqual("HP:0000582", results[3].id)
        self.assertEqual("Upslanted palpebral fissure", results[3].label)
        self.assertTrue(results[3].observed)
        self.assertTrue(results[3].measured)

    def test_mixed_2(self):
        morph_d = {
            'bulbous nasal tip': 'Bulbous nose',
            'prominent lobule of ear': 'Large earlobe',
            'tapering fingers': 'Tapered finger',
            'deeply set eyes': "Deeply set eye" # Note this synonym is not yet in HPO version used for testing
        }
        cell_contents = "Broad forehead, deeply set eyes, ptosis, bulbous nasal tip, micrognathia, prominent lobule of ear, tapering fingers"
        results = self.hpo_cr.parse_cell(cell_contents=cell_contents, custom_d=morph_d)
        # Expect
        # Broad forehead (HP:0000337)
        # Micrognathia (HP:0000347)
        # Bulbous nose (HP:0000414)
        # Deeply set eye (HP:0000490)
        # Ptosis (HP:0000508)
        # Tapered finger (HP:0001182)
        # Large earlobe (HP:0009748)
        self.assertEqual(7, len(results))
        results = sorted(results, key=lambda x: x.id)
        self.assertEqual("HP:0000337", results[0].id)
        self.assertEqual("Broad forehead", results[0].label)
        self.assertTrue(results[0].observed)
        self.assertTrue(results[0].measured)
        self.assertEqual("HP:0000347", results[1].id)
        self.assertEqual("Micrognathia", results[1].label)
        self.assertEqual("HP:0000414", results[2].id)
        self.assertEqual("Bulbous nose", results[2].label)
        self.assertEqual("HP:0000490", results[3].id)
        self.assertEqual("Deeply set eye", results[3].label)
        self.assertEqual("HP:0000508", results[4].id)
        self.assertEqual("Ptosis", results[4].label)
        self.assertEqual("HP:0001182", results[5].id)
        self.assertEqual("Tapered finger", results[5].label)
        self.assertEqual("HP:0009748", results[6].id)
        self.assertEqual("Large earlobe", results[6].label)

    def test_pair_of_custom_concepts(self):
        seizure_d = {'Absence': 'Typical absence seizure',
                    'Infantile spasms': 'Infantile spasms',
                    'GTC': 'Bilateral tonic-clonic seizure',
                    'ESES': 'Status epilepticus'}
        cell_contents = "Absence and GTC"
        results = self.hpo_cr.parse_cell(cell_contents=cell_contents, custom_d=seizure_d)
        ## Expect
        # Typical absence seizure (HP:0011147)
        # Bilateral tonic-clonic seizure (HP:0002069)
        self.assertEqual(2, len(results))
        results = sorted(results, key=lambda x: x.id)
        self.assertEqual("HP:0002069", results[0].id)
        self.assertEqual("Bilateral tonic-clonic seizure", results[0].label)
        self.assertTrue(results[0].observed)
        self.assertTrue(results[0].measured)
        self.assertEqual("HP:0011147", results[1].id)
        self.assertEqual("Typical absence seizure", results[1].label)
        self.assertTrue(results[1].observed)
        self.assertTrue(results[1].measured)

    def test_get_longest_overlapping(self):
        cell_contents = "Cryptorchidism, micropenis, bilateral talipes equinovarus"
        results_longest = self.hpo_cr.parse_cell(cell_contents=cell_contents)
        label_set = {result.label for result in results_longest}
        self.assertIn("Cryptorchidism", label_set, "Could not find Cryptorchidism in results")
        self.assertIn("Micropenis", label_set, "Could not find Micropenis in results")
        self.assertEqual(3, len(label_set))

    def test_initialize_cr(self):
        items = {
            'regression': ["Developmental regression", "HP:0002376"],
            'autism': ['Autism', 'HP:0000717'],
            'hypotonia': ['Hypotonia', 'HP:0001252'],
            'movement disorder': ['Abnormality of movement', 'HP:0100022'],
            'CVI': ['Cerebral visual impairment', 'HP:0100704'],  # CVI stands for Cortical visual impairment HP:0100704
            'seizures': ['Seizure', 'HP:0001250']
        }
        item_column_mapper_d = self.hpo_cr.initialize_simple_column_maps(column_name_to_hpo_label_map=items,
                                                                        observed='yes',
                                                                        excluded='no')
        self.assertEqual(6, len(item_column_mapper_d))


    def test_pica_not_at_boundary(self):
        """
        We do not want Pica HP:0011856 to match with typical
        loss of typical trabecular bony architecture
        """
        cell_contents = "loss of typical trabecular bony architecture"
        results = self.hpo_cr.parse_cell(cell_contents=cell_contents)
        self.assertEqual(0, len(results))

    def test_pica(self):
        """
        This should match
        """
        cell_contents = "Pica is a condition where a person compulsively swallows non-food items."
        results = self.hpo_cr.parse_cell(cell_contents=cell_contents)
        self.assertEqual(1, len(results))






import abc
from typing import List
import pandas as pd
import re

class ColumnMapper(metaclass=abc.ABCMeta):
    """
    Abstract superclass for all Column Mapper classes, each of which applies a specific strategy to extracting HPO terms
    from columns of tables (e.g., supplemental files) representing cohorts of individuals with a given disease.
    """
    def __init__(self, column_name:str) -> None:
        """Constructor

        :param column_name: name of the column in the pandas DataFrame
        :type column_name: str
        """
        self._column_name = column_name

    @abc.abstractmethod
    def map_cell(self, cell_contents) -> List:
        """
        Map a cell to HPO Terms or other data
        """
        pass

    @abc.abstractmethod
    def preview_column(self, df:pd.DataFrame):
        """
        Show a preview of the entire column for review purposes

        :param df: the pandas DataFrame that contains the column (self._column_name)
        :type df: pd.DataFrame
        """
        pass

    @staticmethod
    def is_valid_iso8601(cell_contents):
        """Check for a match with iso8601 age (period)

        returns true for strings such as P6Y, P2M, P42Y1M2W1D etc.

        :returns: true iff the cell_contents represent an iso8601 age
        :rtype: bool
        """
        m = re.match(r"^P(?!$)(\d+Y)?(\d+M)?(\d+W)?(\d+D)?$", cell_contents)
        if m:
            return True
        else:
            return False

    def get_column_name(self):
        return self._column_name





import pandas as pd
from collections import defaultdict
from .constants import Constants



class SexColumnMapper:
    """
    Mapper for a column that indicates the sex of the individual

    :param male_symbol: the synbol for males, e.g. "M", "male", "MALE"
    :type male_symbol: str
    :param female_symbol: the synbol for females, e.g. "F", "female", "FEMALE"
    :type female_symbol: str
    :param column_name: name of the sex column in the original table
    :type column_name: str
    :param other_symbol: symbol for other sex
    :type other_symbol: str
    :param unknown_symbol: symbol for unknown sex
    :type unknown_symbol: str
    :param not_provided: symbol for cases where the sex is not indicated in the original table
    :type not_provided: str
    """


    def __init__(self, male_symbol, female_symbol, column_name, other_symbol=None, unknown_symbol=None, not_provided=False) -> None:
        """
        Constructor
        """
        self._male_symbol = male_symbol
        self._female_symbol = female_symbol
        self._other_symbol = other_symbol
        self._unknown_symbol = unknown_symbol
        if column_name is None:
            raise ValueError("Must provide non-null column_name argument")
        self._column_name = column_name
        self._not_provided = not_provided
        self._erroneous_input_counter = defaultdict(int)


    def map_cell(self, cell_contents) -> str:
        if self._not_provided:
            return Constants.UNKOWN_SEX_SYMBOL
        contents = cell_contents.strip()
        if contents == self._female_symbol:
            return Constants.FEMALE_SYMBOL
        elif contents == self._male_symbol:
            return Constants.MALE_SYMBOL
        elif contents == self._other_symbol:
            return Constants.OTHER_SEX_SYMBOL
        elif contents == self._unknown_symbol:
            return Constants.UNKOWN_SEX_SYMBOL
        else:
            self._erroneous_input_counter[contents] += 1
            return Constants.UNKOWN_SEX_SYMBOL

    def preview_column(self, df:pd.DataFrame):
        if not isinstance(df, pd.DataFrame):
            raise ValueError("df argument must be pandas DataFrame, but was {type(column)}")
        column = df[self.get_column_name()]
        dlist = []
        for _, value in column.items():
            result = self.map_cell(str(value))
            if result is None:
                dlist.append({"original column contents": value, "sex": "n/a"})
            else:
                dlist.append({"original column contents": value, "sex": result})
        return pd.DataFrame(dlist)

    def get_column_name(self):
        return self._column_name

    def has_error(self):
        return len(self._erroneous_input_counter) > 0

    def error_summary(self):
        items = []
        for k, v in self._erroneous_input_counter.items():
            items.append(f"{k} (n={v})")
        return f"Could not parse the following as sex descriptors: {', '.join(items)}"

    @staticmethod
    def not_provided():
        """Create an object for cases where sex is not indicated.
        """
        return SexColumnMapper(male_symbol=Constants.NOT_PROVIDED,
                            female_symbol=Constants.NOT_PROVIDED,
                            not_provided=True,
                            column_name=Constants.NOT_PROVIDED)




import unittest
import os
from src.pyphetools.visualization import HpoCategorySet

from hpotk.ontology.load.obographs import load_ontology

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')


class TestHpoCategorySet(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls._hpo_ontology = load_ontology(HP_JSON_FILENAME)
        cls._category_set = HpoCategorySet(ontology=cls._hpo_ontology)
       
    def check_ataxia(self):
        """
        Ataxia HP:0001251
        """
        cat = self._category_set.get_category("HP:0001251")
        expected = "nervous_system"
        self.assertEqual(expected, cat)

    def check_category_level_term(self):
        """
        Abnormality of the nervous system HP:0000707
        """
        cat = self._category_set.get_category("HP:0000707")
        expected = "nervous_system"
        self.assertEqual(expected, cat)




from .hp_term import HpTerm
from .column_mapper import ColumnMapper
from .hpo_cr import HpoConceptRecognizer
from typing import List
import pandas as pd
import re
from collections import defaultdict


def get_separate_hpos_from_df(df, hpo_cr):
    """Loop through all the cells in a dataframe or series and try to parse each cell as HPO term.
    Useful when the separate HPO terms are in the cells themselves.

    :param df: dataframe with phenotypic data
    :type df: pd.DataFrame
    :param hpo_cr: instance of HpoConceptRecognizer to match HPO term and get label/id
    :type hpo_cr: HpoConceptRecognizer
    :returns: list of lists with the additional HPO terms per individual
    :rtype: List[List[HpTerm]]
    """
    additional_hpos = []

    for i in range(len(df)):
        temp_hpos = []
        for y in range(df.shape[1]):
            hpo_term = hpo_cr.parse_cell(df.iloc[i, y])
            if len(hpo_term) > 0:
                temp_hpos.extend(hpo_term)
        additional_hpos.append(list(set(temp_hpos)))
    return additional_hpos


class SimpleColumnMapper(ColumnMapper):
    """ColumnMapper for columns that contain information about a single phenotypic abnormality only
    :param column_name: name of the column in the pandas DataFrame
    :type column_name: str
    :param hpo_id: HPO  id, e.g., HP:0004321
    :type hpo_id: str
    :param hpo_label: Corresponding term label
    :type hpo_label: str
    :param observed: symbol used in table if the phenotypic feature was observed
    :type observed: str
    :param excluded: symbol used if the feature was excluded
    :type excluded: str
    :param non_measured: symbol used if the feature was not measured or is N/A. Defaults to None, optional
    :type non_measured: str
    """
    def __init__(self, column_name, hpo_id, hpo_label, observed=None, excluded=None, non_measured=None):
        """
        Constructor
        """
        super().__init__(column_name=column_name)
        self._hpo_id = hpo_id
        self._hpo_label = hpo_label
        if observed is None or excluded is None:
            raise ValueError(
                    "Need to provide arguments for both observed and excluded")
        self._observed = observed
        self._excluded = excluded
        self._not_measured = non_measured

    def map_cell(self, cell_contents) -> List[HpTerm]:
        if not isinstance(cell_contents, str):
            raise ValueError(
                f"Error: cell_contents argument ({cell_contents}) must be string but was {type(cell_contents)} -- coerced to string")
        contents = cell_contents.strip()
        # first check if the cell contents represent a valid iso8601, which represents age of onset.
        if ColumnMapper.is_valid_iso8601(contents):
            return [HpTerm(hpo_id=self._hpo_id, label=self._hpo_label, onset=contents)]
        if contents in self._observed:
            return [HpTerm(hpo_id=self._hpo_id, label=self._hpo_label)]
        elif contents in self._excluded:
            return [HpTerm(hpo_id=self._hpo_id, label=self._hpo_label, observed=False)]
        else:
            return [HpTerm(hpo_id=self._hpo_id, label=self._hpo_label, measured=False)]

    def preview_column(self, df:pd.DataFrame) -> pd.DataFrame:
        if not isinstance(df, pd.DataFrame):
            raise ValueError("df argument must be pandas DataFrame, but was {type(column)}")
        column = df[self._column_name]
        mapping_counter = defaultdict(int)
        for _, value in column.items():
            cell_contents = str(value)
            value = self.map_cell(cell_contents)
            hpterm = value[0]
            mapped = f"original value: \"{cell_contents}\" -> HP: {hpterm.hpo_term_and_id} ({hpterm.display_value})"
            mapping_counter[mapped] += 1
        dlist = []
        for k, v in mapping_counter.items():
            d = {"mapping": k, "count": str(v)}
            dlist.append(d)
        return pd.DataFrame(dlist)

class SimpleColumnMapperGenerator:
    """Convenience tool to provide mappings automatically

    Try to map the columns in a dataframe by matching the name of the column to correct HPO term.
    This class can be used to generate SimpleColumn mappers for exact matches found in the columns names.

    :param df: dataframe with phenotypic data
    :type df: pd.DataFrame
    :param observed: symbol used in table if the phenotypic feature was observed
    :type observed: str
    :param excluded: symbol used if the feature was excluded
    :type excluded: str
    :param hpo_cr: instance of HpoConceptRecognizer to match HPO term and get label/id
    :type hpo_cr: HpoConceptRecognizer
    """
    def __init__(self, df:pd.DataFrame, observed:str, excluded:str, hpo_cr:HpoConceptRecognizer) -> None:
        """
        Constructor
        """
        self._df = df
        self._observed = observed
        self._excluded = excluded
        self._hpo_cr = hpo_cr
        self._mapped_columns = []
        self._unmapped_columns = []
        self._error_messages = []


    def try_mapping_columns(self) -> List[ColumnMapper]:
        """As a side effect, this class initializes three lists of mapped, unmapped, and error columns

        :returns: A dictionary with successfully mapped columns
        :rtype: Dict[str,ColumnMapper]
        """
        simple_mapper_list = list()
        hpo_id_re = r"(HP:\d+)"
        for col in self._df.columns:
            colname = str(col)
            result = re.search(r"(HP:\d+)", colname)
            if self._hpo_cr.contains_term_label(colname):
                hpo_term_list = self._hpo_cr.parse_cell(colname)
                hpo_term = hpo_term_list[0]
                scm = SimpleColumnMapper(column_name=colname,
                                        hpo_id=hpo_term.id,
                                        hpo_label=hpo_term.label,
                                        observed=self._observed,
                                        excluded=self._excluded)
                simple_mapper_list.append(scm)
            elif result:
                hpo_id = result.group(1)
                if self._hpo_cr.contains_term(hpo_id):
                    hterm = self._hpo_cr.get_term_from_id(hpo_id)
                    scm = SimpleColumnMapper(column_name=colname,
                                            hpo_id=hterm.id,
                                            hpo_label=hterm.label,
                                            observed=self._observed,
                                            excluded=self._excluded)
                    simple_mapper_list.append(scm)
                else:
                    self._unmapped_columns.append(colname)
            else:
                self._unmapped_columns.append(colname)
        self._mapped_columns = [scm.get_column_name() for scm in simple_mapper_list]
        return simple_mapper_list


    def get_unmapped_columns(self):
        """
        :returns: A list of names of the columns that could not be mapped
        :rtype: List[str]
        """
        return self._unmapped_columns

    def get_mapped_columns(self) -> List[str]:
        """
        :returns: A list of names of the columns that were mapped
        :rtype: List[str]
        """
        return self._mapped_columns

    def to_html(self):
        """create an HTML table with names of mapped and unmapped columns
        """
        table_items = []
        table_items.append('<table style="border: 2px solid black;">\n')
        table_items.append("""<tr>
            <th>Result</th>
            <th>Columns</th>
        </tr>
        """)
        mapped_str = "; ".join(self._mapped_columns)
        unmapped_str = "; ".join([f"<q>{colname}</q>" for colname in self._unmapped_columns])
        def two_item_table_row(cell1, cell2):
            return f"<tr><td>{cell1}</td><td>{cell2}</td></tr>"
        table_items.append(two_item_table_row("Mapped", mapped_str))
        table_items.append(two_item_table_row("Unmapped", unmapped_str))
        table_items.append('</table>\n') # close table content
        return "\n".join(table_items)



import unittest

from src.pyphetools.creation import HpTerm, HpTermBuilder


class TestHpTerm(unittest.TestCase):

    def test_excluded(self):
        hpo_term = HpTerm(hpo_id="HP:0001250", label="test label")
        self.assertEqual("HP:0001250", hpo_term.id)
        self.assertEqual("test label", hpo_term.label)
        self.assertTrue(hpo_term.observed)
        self.assertTrue(hpo_term.measured)
        hpo_term.excluded()
        self.assertFalse(hpo_term.observed)


    def test_create_phenotypic_feature(self):
        """test creation of GA4GH PhenotypicFeature object with onset and resolution
        """
        hpo_term = HpTerm(hpo_id="HP:0001250", label="test label", onset="P2M", resolution="P6M")
        self.assertEqual("HP:0001250", hpo_term.id)
        self.assertEqual("test label", hpo_term.label)
        self.assertEqual("P2M", hpo_term.onset)
        self.assertEqual("P6M", hpo_term.resolution)
        self.assertTrue(hpo_term.observed)
        self.assertTrue(hpo_term.measured)
        pfeat = hpo_term.to_ga4gh_phenotypic_feature()
        self.assertIsNotNone(pfeat)
        pfterm = pfeat.type
        self.assertEqual("HP:0001250", pfterm.id)
        self.assertEqual("test label", pfterm.label)
        self.assertEqual("P2M", pfeat.onset.age.iso8601duration)
        self.assertEqual("P6M", pfeat.resolution.age.iso8601duration)
        self.assertFalse(pfeat.excluded)


    def test_fetal_onset(self):
        hpterm = HpTermBuilder(hpo_id="HP:0010942", hpo_label="Echogenic intracardiac focus").fetal_onset().build()
        self.assertEqual("HP:0010942", hpterm.id)
        self.assertEqual("Echogenic intracardiac focus", hpterm.label)
        self.assertEqual('NOT_PROVIDED', hpterm.onset)
        self.assertIsNotNone(hpterm.onset_term)
        onset_term = hpterm.onset_term
        self.assertTrue(isinstance(onset_term, list))
        self.assertEqual(2, len(onset_term))
        # Fetal onset HP:0011461
        self.assertEqual("Fetal onset", onset_term[0])
        self.assertEqual("HP:0011461", onset_term[1])
        phenotypic_feature = hpterm.to_ga4gh_phenotypic_feature()
        self.assertIsNotNone(phenotypic_feature.onset)
        self.assertIsNotNone(phenotypic_feature.onset.ontology_class)
        oclass = phenotypic_feature.onset.ontology_class
        self.assertEqual("HP:0011461", oclass.id)
        self.assertEqual("Fetal onset", oclass.label)

    def test_late_onset(self):
        hpterm = HpTermBuilder(hpo_id="HP:0000726", hpo_label="Dementia").late_onset().build()
        self.assertEqual("HP:0000726", hpterm.id)
        self.assertEqual("Dementia", hpterm.label)
        self.assertEqual('NOT_PROVIDED', hpterm.onset)
        self.assertIsNotNone(hpterm.onset_term)
        onset_term = hpterm.onset_term
        self.assertTrue(isinstance(onset_term, list))
        self.assertEqual(2, len(onset_term))
        # Late onset HP:0003584
        self.assertEqual("Late onset", onset_term[0])
        self.assertEqual("HP:0003584", onset_term[1])
        phenotypic_feature = hpterm.to_ga4gh_phenotypic_feature()
        self.assertIsNotNone(phenotypic_feature.onset)
        self.assertIsNotNone(phenotypic_feature.onset.ontology_class)
        oclass = phenotypic_feature.onset.ontology_class
        self.assertEqual("HP:0003584", oclass.id)
        self.assertEqual("Late onset", oclass.label)

    def test_get_exception_if_add_both_iso_and_term(self):
        """We expect a ValueError because one is not allowed to add both an iso age string and an HPO onset class
        """
        with self.assertRaises(ValueError) as context:
            hpterm = HpTermBuilder(hpo_id="HP:0000726", hpo_label="Dementia").late_onset().iso8601_onset("P65Y").build()
        self.assertTrue('Both onset and onsetTerm were set but only a maximum of one of them may be not None' in str(context.exception))






import pandas as pd
from math import isnan
from typing import Dict, List

from .abstract_encoder import AbstractEncoder
from .age_column_mapper import AgeColumnMapper
from .citation import Citation
from .constants import Constants
from .disease import Disease
from .hpo_cr import HpoConceptRecognizer
from .individual import Individual
from .sex_column_mapper import SexColumnMapper
from .variant_column_mapper import VariantColumnMapper


class CohortEncoder(AbstractEncoder):
    """Map a table of data to Individual/GA4GH Phenopacket Schema objects

    Encode a cohort of individuals with clinical data in a table as a collection of GA4GH Phenopackets
    This classes uses a collection of ColumnMapper objects to map a table using the
    get_individuals or output_phenopackets methods.

    The column_mapper_d is a dictionary with key=column names, and value=Mapper objects. These mappers are responsible
    for mapping HPO terms. The agemapper and the sexmapper are specialized for the respective columns. The
    variant mapper is useful if there is a single variant column that is all HGVS or structural variants. In some
    cases, it is preferable to use the variant_dictionary, which has key=string (cell contents) and value=Hgvs or
    StructuralVariant object.

    :param df: tabular data abotu a cohort
    :type df: pd.DataFrame
    :param hpo_cr: HpoConceptRecognizer for text mining
    :type hpo_cr: pyphetools.creation.HpoConceptRecognizer
    :param column_mapper_list: list of ColumnMapper objects
    :type column_mapper_list: List[pyphetools.creation.ColumnMapper]
    :param individual_column_name: label of column with individual/proband/patient identifier
    :type individual_column_name: str
    :param metadata: GA4GH MetaData object
    :type metadata: PPkt.MetaData
    :param agemapper:Mapper for the Age column. Defaults to AgeColumnMapper.not_provided()
    :type agemapper: pyphetools.creation.AgeColumnMapper
    :param sexmapper: Mapper for the Sex column. Defaults to SexColumnMapper.not_provided().
    :type sexmapper: pyphetools.creation.SexColumnMapper
    :param variant_mapper: column mapper for HGVS-encoded variant column.
    :type variant_mapper: pyphetools.creation.VariantColumnMapper
    :raises: ValueError - several of the input arguments are checked.
    """

    def __init__(self,
                df,
                hpo_cr,
                column_mapper_list,
                individual_column_name,
                metadata,
                agemapper=AgeColumnMapper.not_provided(),
                sexmapper=SexColumnMapper.not_provided(),
                variant_mapper=None,
                delimiter=None):
        """Constructor
        """
        super().__init__(metadata=metadata)
        if not isinstance(hpo_cr, HpoConceptRecognizer):
            raise ValueError(
                "concept_recognizer argument must be HpoConceptRecognizer but was {type(concept_recognizer)}")
        self._hpo_concept_recognizer = hpo_cr
        if not isinstance(df, pd.DataFrame):
            raise ValueError(f"df argument must be pandas data frame but was {type(df)}")
        if not isinstance(column_mapper_list, list):
            raise ValueError(f"column_mapper_list argument must be a list but was {type(column_mapper_list)}")
        if not isinstance(individual_column_name, str):
            raise ValueError(f"individual_column_name argument must be a string but was {type(individual_column_name)}")
        if variant_mapper is not None and not isinstance(variant_mapper, VariantColumnMapper):
            raise ValueError(f"variant_mapper argument must be VariantColumnMapper but was {type(variant_mapper)}")
        self._df = df.astype(str)
        self._column_mapper_list = column_mapper_list
        self._id_column_name = individual_column_name
        self._age_mapper = agemapper
        self._sex_mapper = sexmapper
        self._disease = None
        self._variant_mapper = variant_mapper
        self._disease_dictionary = None
        self._delimiter = delimiter
        ontology = hpo_cr.get_hpo_ontology()
        if ontology is None:
            raise ValueError("ontology cannot be None")

    def preview_dataframe(self):
        """
        Generate a dataframe with a preview of the parsed contents

        :returns: a DataFrame representing the cohort to check results
        :rtype: pd.DataFrame
        """
        df = self._df.reset_index()  # make sure indexes pair with number of rows
        individuals = []
        age_column_name = self._age_mapper.get_column_name()
        sex_column_name = self._sex_mapper.get_column_name()
        for index, row in df.iterrows():
            individual_id = row[self._id_column_name]
            if age_column_name == Constants.NOT_PROVIDED:
                age = Constants.NOT_PROVIDED
            else:
                age_cell_contents = row[age_column_name]
                age = self._age_mapper.map_cell(age_cell_contents)
            sex_cell_contents = row[sex_column_name]
            sex = self._sex_mapper.map_cell(sex_cell_contents)
            hpo_terms = []
            for column_mapper in self.__column_mapper_list:
                column_name = column_mapper.get_column_name()
                cell_contents = row[column_name]
                # Empty cells are often represented as float non-a-number by Pandas
                if isinstance(cell_contents, float) and isnan(cell_contents):
                    continue
                terms = column_mapper.map_cell(row[column_name])
                hpo_terms.extend(terms)
            hpo_string = "\n".join([h.to_string() for h in hpo_terms])
            d = {'id': individual_id,
                'sex': sex,
                'age': age,
                'phenotypic features': hpo_string}
            individuals.append(d)
        df = pd.DataFrame(individuals)
        return df.set_index('id')

    def set_disease(self, disease:Disease):
        """Set the disease diagnosis for all patients in the cohort

        If all patients in the cohort have the same disease we can set it with this method
        :param disease: Disease diagnosis for the individuals in this cohort
        :type disease: Disease
        """
        self._disease = disease
        self._disease_dictionary = None

    def set_disease_dictionary(self, disease_d:Dict[str, Disease]):
        """Set the dictionary of disease ontology terms

        For tables with multiple different diseases, we provide a dictionary that has as key
        the string used in the original table and as value
        """
        self._disease_dictionary = disease_d
        self._disease = None

    def get_individuals(self) -> List[Individual]:
        """Get a list of all Individual objects in the cohort

        :returns: a list of all Individual objects in the cohort
        :rtype: List[Individual]
        """
        # make sure indexes pair with number of rows, if needed
        if not self._df.index.name in self._df.columns:
            df = self._df.reset_index()
        individuals = []
        age_column_name = self._age_mapper.get_column_name()
        sex_column_name = self._sex_mapper.get_column_name()
        if self._variant_mapper is None:
            variant_colname = None
            genotype_colname = None
        else:
            variant_colname = self._variant_mapper.get_variant_column_name()
            genotype_colname = self._variant_mapper.get_genotype_colname()
        for index, row in df.iterrows():
            individual_id = row[self._id_column_name]
            if age_column_name == Constants.NOT_PROVIDED:
                age = Constants.NOT_PROVIDED
            else:
                age_cell_contents = row[age_column_name]
                try:
                    age = self._age_mapper.map_cell(age_cell_contents)
                except Exception as ee:
                    print(f"Warning: Could not parse age {ee}. Setting age to \"not provided\"")
                    age = Constants.NOT_PROVIDED
            if sex_column_name == Constants.NOT_PROVIDED:
                sex = self._sex_mapper.map_cell(Constants.NOT_PROVIDED)
            else:
                sex_cell_contents = row[sex_column_name]
                sex = self._sex_mapper.map_cell(sex_cell_contents)
            hpo_terms = []
            for column_mapper in self._column_mapper_list:
                column_name = column_mapper.get_column_name()
                if column_name not in df.columns:
                    raise ValueError(f"Did not find column name '{column_name}' in dataframe -- check spelling!")
                cell_contents = row[column_name]
                # Empty cells are often represented as float non-a-number by Pandas
                if isinstance(cell_contents, float) and isnan(cell_contents):
                    continue
                terms = column_mapper.map_cell(cell_contents)
                hpo_terms.extend(terms)
            if variant_colname is not None:
                variant_cell_contents = row[variant_colname]
                if genotype_colname is not None:
                    genotype_cell_contents = row[genotype_colname]
                else:
                    genotype_cell_contents = None
                if self._variant_mapper is not None:
                    interpretation_list = self._variant_mapper.map_cell(variant_cell_contents, genotype_cell_contents)
                else:
                    interpretation_list = []
            else:
                interpretation_list = []
            if self._disease_dictionary is not None and self._disease is None:
                if individual_id not in self._disease_dictionary:
                    raise ValueError(f"Could not find disease link for {individual_id}")
                disease = self._disease_dictionary.get(individual_id)
                indi = Individual(individual_id=individual_id,
                                sex=sex,
                                age=age,
                                hpo_terms=hpo_terms,
                                citation=self._metadata.get_citation(),
                                interpretation_list=interpretation_list,
                                disease=disease)
            elif self._disease_dictionary is None and self._disease is not None:
                indi = Individual(individual_id=individual_id,
                                sex=sex,
                                age=age,
                                hpo_terms=hpo_terms,
                                citation=self._metadata.get_citation(),
                                interpretation_list=interpretation_list,
                                disease=self._disease)
            else:
                raise ValueError(f"Could not find disease data for '{individual_id}'")
            individuals.append(indi)
        if self._age_mapper.has_error():
            print(self._age_mapper.error_summary())
        if self._sex_mapper.has_error():
            print(self._sex_mapper.error_summary())
        return individuals



import unittest
import os
import pandas as pd
from src.pyphetools.creation import HpoParser, CaseEncoder, MetaData, Citation
HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')
import phenopackets as PPkt

# The API requires us to pass a column name but the column name will not be used in the tests
TEST_COLUMN = "test"


class TestCaseParse(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        hpparser = HpoParser(hpo_json_file=HP_JSON_FILENAME)
        cls._hpo_cr = hpparser.get_hpo_concept_recognizer()
        metadata = MetaData(created_by="ORCID:0000-0002-0736-9199")
        metadata.default_versions_with_hpo(version="2022-05-05")
        age_of_last_examination = "P4Y11M"
        sex = "female"
        citation = Citation(pmid="PMID:123", title="example title")
        cls._parser = CaseEncoder(hpo_cr=cls._hpo_cr,
                                individual_id="A",
                                citation=citation,
                                age_at_last_exam=age_of_last_examination,
                                sex=sex,
                                metadata=metadata.to_ga4gh())

    def test_chief_complaint(self):
        """
        Expect to get
        HP:0000365 	Hearing impairment 	True 	True
        HP:0001742 	Nasal congestion 	True 	True
        HP:0025267 	Snoring
        """
        vignette = "A 17-mo-old boy presented with progressive nasal obstruction, snoring and hearing loss symptoms when referred to the hospital."
        results = self._parser.add_vignette(vignette=vignette)
        self.assertEqual(3, len(results))
        self.assertTrue(isinstance(results, pd.DataFrame))
        tid = results.loc[(results['id'] == 'HP:0000365')]['id'].values[0]
        self.assertEqual("HP:0000365", tid)

    def test_excluded(self):
        """
        Currently, our text mining does not attempt to detect negation, but users can enter a set of term labels that are excluded
        In this case, He had no nasal obstruction should lead to excluded/Nasal congestion
        """
        vignette = "He had no nasal obstruction and he stated that his smell sensation was intact."
        excluded = set()
        excluded.add("Nasal congestion")
        results = self._parser.add_vignette(vignette=vignette, excluded_terms=excluded)
        self.assertEqual(1, len(results))
        tid = results.loc[(results['id'] == 'HP:0001742')]['id'].values[0]
        self.assertEqual("HP:0001742", tid)
        label = results.loc[(results['id'] == 'HP:0001742')]['label'].values[0]
        self.assertEqual("Nasal congestion", label)
        observed = results.loc[(results['id'] == 'HP:0001742')]['observed'].values[0]
        self.assertFalse(observed)
        measured = results.loc[(results['id'] == 'HP:0001742')]['measured'].values[0]
        self.assertTrue(measured)

    def test_excluded2(self):
        """
        The goal if this test is to ensure that 'Seizure' is annotated as 'excluded'
        We label the word as false positive and it should just be skipped
        """
        vignette = "The patient is not on seizure medication at this time."
        excluded = set()
        false_positive = set()
        false_positive.add("seizure")
        metadata = MetaData(created_by="ORCID:0000-0002-0736-9199")
        metadata.default_versions_with_hpo(version="2022-05-05")
        citation = Citation(pmid="PMID:1", title="excluded")
        encoder = CaseEncoder(hpo_cr=self._hpo_cr, individual_id="id", metadata=metadata.to_ga4gh(), citation=citation)
        df = encoder.add_vignette(vignette=vignette, excluded_terms=excluded, false_positive=false_positive)
        self.assertEqual(0, len(df))



    def test_phenopacket_id(self):
        """
        Test that we construct the phenopacket ID to include the PMID and the individual id
        """
        ppkt = self._parser.get_phenopacket()
        expected_ppkt_id = "PMID_123_A"
        self.assertEqual(expected_ppkt_id, ppkt.id)


    def test_age_last_exam(self):
        individual = self._parser.get_individual()
        self.assertIsNotNone(individual)
        self.assertIsNotNone(individual.age)
        expected_age = "P4Y11M"
        self.assertEqual(expected_age, individual.age)

    def test_sex(self):
        ppkt = self._parser.get_phenopacket()
        self.assertIsNotNone(ppkt.subject)
        self.assertIsNotNone(ppkt.subject.sex)
        expected_sex = PPkt.Sex.FEMALE
        self.assertEqual(expected_sex, ppkt.subject.sex)





import unittest
from src.pyphetools.creation import AgeOfDeathColumnMapper, Individual, MetaData
import phenopackets as PPKt



class TestVitalStatus(unittest.TestCase):

    def test_vs1(self):
        aod_d = {
            "7 months": "P7M",
            "10 months": "P10M",
        }
        mapper = AgeOfDeathColumnMapper(column_name='test', string_to_iso_d=aod_d)
        vstatus = mapper.map_cell_to_vital_status("not there")
        self.assertIsNone(vstatus)
        vstatus = mapper.map_cell_to_vital_status("7 months")
        self.assertIsNotNone(vstatus)
        self.assertTrue(isinstance(vstatus, PPKt.VitalStatus))
        self.assertEqual(PPKt.VitalStatus.DECEASED, vstatus.status)
        self.assertEqual("P7M", vstatus.time_of_death.age.iso8601duration)

    def test_vs2(self):
        aod_d = {
            "7 months": "P7M",
            "10 months": "P10M",
        }
        mapper = AgeOfDeathColumnMapper(column_name='test', string_to_iso_d=aod_d)
        vstatus = mapper.map_cell_to_vital_status("10 months")
        self.assertIsNotNone(vstatus)
        self.assertTrue(isinstance(vstatus, PPKt.VitalStatus))
        self.assertEqual(PPKt.VitalStatus.DECEASED, vstatus.status)
        self.assertEqual("P10M", vstatus.time_of_death.age.iso8601duration)

    def test_individual_with_vs(self):
        """Test that we can add a VitalStatus to the phenopacket.
        """
        aod_d = {
            "7 months": "P7M",
            "10 months": "P10M",
        }
        mapper = AgeOfDeathColumnMapper(column_name='test', string_to_iso_d=aod_d)
        vstatus = mapper.map_cell_to_vital_status("10 months")
        i = Individual(individual_id="test")
        i.set_vital_status(vstatus=vstatus)
        ## needed for API
        hpo_version = "fake.version"
        metadata = MetaData(created_by="ORCID:0000-0002-0736-9199")
        metadata.default_versions_with_hpo(version=hpo_version)
        ppkt = i.to_ga4gh_phenopacket(metadata)
        aod = ppkt.subject.vital_status.time_of_death.age.iso8601duration
        self.assertEqual("P10M", aod)
        print(i)





class Citation:
    """encapsulate information about a citation that we add to the metadata for display

    :param pmid: PubMed identifier for the publication in which this individual was described (e.g. PMID:321..).
    :type pmid: str
    :param title: Title of the publication in which this individual was described.
    :type title: str
    """

    def __init__(self, pmid, title) -> None:
        self._pmid = pmid
        self._title = title


    @property
    def pmid(self):
        return self._pmid

    @property
    def title(self):
        return self._title





class Disease:
    """
    Simple data object to hold ontology id/label for a disease diagnosis

    :param disease_id: a CURIE such as OMIM:600324
    :type disease_id: str
    :param disease_label: the name of the disease
    :type disease_label: str
    """

    def __init__(self, disease_id:str, disease_label):
        self._id = disease_id
        self._label = disease_label


    @property
    def id(self):
        return self._id

    @property
    def label(self):
        return self._label

    def __hash__(self):
        return hash((self._id, self._label))

    def __eq__(self, other):
        return (self._id, self._label) == (other._id, other._label)

    def __repr__(self):
        return f'{self._label} ({self._id})'



import unittest
from src.pyphetools.creation import AgeIsoFormater



class TestAgeIsoFormater(unittest.TestCase):

    def test_basic1(self):
        iso_age = AgeIsoFormater.to_string(y=2, m=3, d=5)
        self.assertEqual("P2Y3M5D", iso_age)

    def test_basic2(self):
        """
        test that 13 months are normalized to 1 year 1 month
        """
        iso_age = AgeIsoFormater.to_string(y=42, m=13, d=5)
        self.assertEqual("P43Y1M5D", iso_age)

    def test_5m(self):
        iso_age = AgeIsoFormater.from_numerical_month(5)
        self.assertEqual("P5M", iso_age)

    def test_15d(self):
        iso_age = AgeIsoFormater.from_numerical_month(0.5)
        self.assertEqual("P15D", iso_age)

    def test_24d(self):
        iso_age = AgeIsoFormater.from_numerical_month(0.8)
        self.assertEqual("P24D", iso_age)

    def test_12m(self):
        iso_age = AgeIsoFormater.from_numerical_month(12)
        self.assertEqual("P1Y", iso_age)

    def test_16m(self):
        iso_age = AgeIsoFormater.from_numerical_month(16)
        self.assertEqual("P1Y4M", iso_age)

    def test_na(self):
        """
        Test we return NOT_PROVIDED (from the Contants class) if we cannot parse the cell contents
        """
        iso_age = AgeIsoFormater.from_numerical_month("n.a.")
        self.assertEqual("NOT_PROVIDED", iso_age)

    def test_none(self):
        """
        Test we return NOT_PROVIDED (from the Contants class) if we cannot parse the cell contents
        """
        iso_age = AgeIsoFormater.from_numerical_month(None)
        self.assertEqual("NOT_PROVIDED", iso_age)

    def test_nan(self):
        """
        Test we return NOT_PROVIDED (from the Contants class) if we cannot parse the cell contents
        """
        iso_age = AgeIsoFormater.from_numerical_month(float("nan"))
        self.assertEqual("NOT_PROVIDED", iso_age)

    def test_newborn(self):
        iso_age = AgeIsoFormater.from_numerical_month(0)
        self.assertEqual("P0D", iso_age)






from math import isnan
from typing import List, Dict
import os
import re
from google.protobuf.json_format import MessageToJson
import phenopackets as PPKt
import pandas as pd

from . import Individual
from .abstract_encoder import AbstractEncoder
from .age_column_mapper import AgeColumnMapper
from .age_of_death_mapper import AgeOfDeathColumnMapper
from .citation import Citation
from .constants import Constants
from .hpo_cr import HpoConceptRecognizer
from .metadata import MetaData
from .sex_column_mapper import SexColumnMapper
from .variant_column_mapper import VariantColumnMapper


class MixedCohortEncoder(AbstractEncoder):

    HPO_VERSION = None

    """Map a table of data to Individual/GA4GH Phenopacket Schema objects. This class should be used instead of CohortMapper
       if the table has different diseases or different pubmed references for each row.

        Encode a cohort of individuals with clinical data in a table as a collection of GA4GH Phenopackets
        This classes uses a collection of ColumnMapper objects to map a table using the
        get_individuals or output_phenopackets methods.

        The column_mapper_d is a dictionary with key=column names, and value=Mapper objects. These mappers are responsible
        for mapping HPO terms. The agemapper and the sexmapper are specialized for the respective columns. The
        variant mapper is useful if there is a single variant column that is all HGVS or structural variants. In some
        cases, it is preferable to use the variant_dictionary, which has key=string (cell contents) and value=Hgvs or
        StructuralVariant object.

        :param df: tabular data abotu a cohort
        :type df: pd.DataFrame
        :param hpo_cr: HpoConceptRecognizer for text mining
        :type hpo_cr: pyphetools.creation.HpoConceptRecognizer
        :param column_mapper_list: list of ColumnMapper objects
        :type column_mapper_list: List[pyphetools.creation.ColumnMapper]
        :param individual_column_name: label of column with individual/proband/patient identifier
        :type individual_column_name: str
        :param disease_id_mapper: label of column with disease id
        :type disease_id_mapper: str
        :param metadata: GA4GH MetaData object
        :type metadata: PPkt.MetaData
        :param variant_mapper: column mapper for HGVS-encoded variant column. Defaults to None.
        :type variant_mapper: pyphetools.creation.VariantColumnMapper
        :param agemapper:Mapper for the Age column. Defaults to AgeColumnMapper.not_provided()
        :type agemapper: pyphetools.creation.AgeColumnMapper
        :param sexmapper: Mapper for the Sex column. Defaults to SexColumnMapper.not_provided().
        :type sexmapper: pyphetools.creation.SexColumnMapper
        :param pmid_column: name of column with PubMed identifier for the row
        :type pmid_column: str
        :raises: ValueError - several of the input arguments are checked.
        """
    def __init__(self,
                df,
                hpo_cr,
                column_mapper_list,
                individual_column_name,
                disease_id_mapper,
                metadata,
                pmid_column:str,
                title_column:str=None,
                variant_mapper:VariantColumnMapper=None,
                agemapper=AgeColumnMapper.not_provided(),
                sexmapper=SexColumnMapper.not_provided(),
                age_of_death_mapper:AgeOfDeathColumnMapper=None,
                delimiter=None):
        """Constructor
        """
        super().__init__(metadata=metadata)
        if not isinstance(hpo_cr, HpoConceptRecognizer):
            raise ValueError(
                "concept_recognizer argument must be HpoConceptRecognizer but was {type(concept_recognizer)}")
        self._hpo_concept_recognizer = hpo_cr
        if not isinstance(df, pd.DataFrame):
            raise ValueError(f"df argument must be pandas data frame but was {type(df)}")
        if not isinstance(column_mapper_list, list):
            raise ValueError(f"column_mapper_list argument must be a list but was {type(column_mapper_list)}")
        if not isinstance(individual_column_name, str):
            raise ValueError(f"individual_column_name argument must be a string but was {type(individual_column_name)}")
        self._df = df.astype(str)
        self._column_mapper_list = column_mapper_list
        self._id_column_name = individual_column_name
        self._age_mapper = agemapper
        self._sex_mapper = sexmapper
        self._age_of_death_mapper = age_of_death_mapper
        self._disease_id_mapper = disease_id_mapper
        self._pmid_column = pmid_column
        self._title_column = title_column
        self._variant_mapper = variant_mapper
        self._delimiter = delimiter
        MixedCohortEncoder.HPO_VERSION = hpo_cr.get_hpo_ontology().version


    def get_individuals(self) -> List[Individual]:
        """Get a list of all Individual objects in the cohort

        :returns: a list of all Individual objects in the cohort
        :rtype: List[Individual]
        """
        df = self._df.reset_index()  # make sure indexes pair with number of rows
        individuals = []
        age_column_name = self._age_mapper.get_column_name()
        sex_column_name = self._sex_mapper.get_column_name()
        if self._age_of_death_mapper is not None:
            age_of_death_column_name = self._age_of_death_mapper.column_name
        else:
            age_of_death_column_name = None
        disease_column_name = self._disease_id_mapper.get_column_name()
        if self._variant_mapper is None:
            variant_colname = None
            genotype_colname = None
        else:
            variant_colname = self._variant_mapper.get_variant_column_name()
            genotype_colname = self._variant_mapper.get_genotype_colname()
        for index, row in df.iterrows():
            individual_id = row[self._id_column_name]
            if age_column_name == Constants.NOT_PROVIDED:
                age = Constants.NOT_PROVIDED
            else:
                age_cell_contents = row[age_column_name]
                try:
                    age = self._age_mapper.map_cell(age_cell_contents)
                except Exception as ee:
                    print(f"Warning: Could not parse age {ee}. Setting age to \"not provided\"")
                    age = Constants.NOT_PROVIDED
            if sex_column_name == Constants.NOT_PROVIDED:
                sex = self._sex_mapper.map_cell(Constants.NOT_PROVIDED)
            else:
                sex_cell_contents = row[sex_column_name]
                sex = self._sex_mapper.map_cell(sex_cell_contents)
            if age_of_death_column_name is not None:
                age_of_death_contents = row[age_of_death_column_name]
                vstatus = self._age_of_death_mapper.map_cell_to_vital_status(age_of_death_contents)
            else:
                vstatus = None
            pmid = row[self._pmid_column]
            if self._title_column is not None:
                title = row[self._title_column]
            else:
                title = None
            if pmid is None:
                raise ValueError(f"Could not get PMID (required) for individual {individual_id}")
            if title is None:
                raise ValueError(f"Could not get title (required) for individual {individual_id}")
            hpo_terms = []
            for column_mapper in self._column_mapper_list:
                column_name = column_mapper.get_column_name()
                if column_name not in df.columns:
                    raise ValueError(f"Did not find column name '{column_name}' in dataframe -- check spelling!")
                cell_contents = row[column_name]
                # Empty cells are often represented as float non-a-number by Pandas
                if isinstance(cell_contents, float) and isnan(cell_contents):
                    continue
                terms = column_mapper.map_cell(cell_contents)
                hpo_terms.extend(terms)
            if variant_colname is not None:
                variant_cell_contents = row[variant_colname]
                if genotype_colname is not None:
                    genotype_cell_contents = row[genotype_colname]
                else:
                    genotype_cell_contents = None
                if self._variant_mapper is not None:
                    interpretation_list = self._variant_mapper.map_cell(variant_cell_contents, genotype_cell_contents)
                else:
                    interpretation_list = []
            else:
                interpretation_list = []
            disease_cell_contents = row[disease_column_name]
            if disease_cell_contents is None:
                raise ValueError(f"Could not extract disease identifier for row {row}")
            disease = self._disease_id_mapper.map_cell(disease_cell_contents)
            cite = Citation(pmid=pmid, title=title)
            indi = Individual(individual_id=individual_id,
                                sex=sex,
                                age=age,
                                hpo_terms=hpo_terms,
                                citation=cite,
                                interpretation_list=interpretation_list,
                                disease=disease)
            if vstatus is not None:
                indi.set_vital_status(vstatus)
            individuals.append(indi)
        return individuals

    @staticmethod
    def output_individuals_as_phenopackets(individual_list:List[Individual], created_by=None, outdir="phenopackets"):
        """write a list of Individual objects to file in GA4GH Phenopacket format


        :param outdir: Path to output directory. Defaults to "phenopackets". Created if not exists.
        :type outdir: str
        """
        if os.path.isfile(outdir):
            raise ValueError(f"Attempt to create directory with name of existing file {outdir}")
        if not os.path.isdir(outdir):
            os.makedirs(outdir)
        written = 0

        if created_by is None:
            created_by = 'pyphetools'
        for individual in individual_list:
            cite = individual._citation
            metadata = MetaData(created_by=created_by, citation=cite)
            metadata.default_versions_with_hpo(MixedCohortEncoder.HPO_VERSION)
            phenopckt = individual.to_ga4gh_phenopacket(metadata=metadata)
            json_string = MessageToJson(phenopckt)
            pmid = cite.pmid
            if pmid is None:
                fname = "phenopacket_" + individual.id
            else:
                pmid = pmid.replace(" ", "").replace(":", "_")
                fname = pmid + "_" + individual.id
            fname = re.sub('[^A-Za-z0-9_-]', '', fname)  # remove any illegal characters from filename
            fname = fname.replace(" ", "_") + ".json"
            outpth = os.path.join(outdir, fname)
            with open(outpth, "wt") as fh:
                fh.write(json_string)
                written += 1
        print(f"We output {written} GA4GH phenopackets to the directory {outdir}")



import time
from collections import defaultdict

import phenopackets as PPKt
from google import protobuf

from .citation import Citation


class Resource:
    def __init__(self, resource_id, name, namespace_prefix, iriprefix, url, version) -> None:
        self._id = resource_id
        self._name = name
        self._namespace_prefix = namespace_prefix
        self._iri_prefix = iriprefix
        self._url = url
        self._version = version

    def to_ga4gh(self):
        pass

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def namespace_prefix(self):
        return self._namespace_prefix

    @property
    def iri_prefix(self):
        return self._iri_prefix

    @property
    def url(self):
        return self._url

    @property
    def version(self):
        return self._version


# current versions of ontologies used for these phenopackets
# if the user does not specify versions, these versions will be
# used.
# for our purposes, the one version that will change a lot is that
# of the HPO and users can get that from the HPO file and this should be
# provided in the constructor of metadata

default_versions = {
    'geno': '2022-03-05',
    'hgnc': '06/01/23',
    'omim': 'January 4, 2023',
    'mondo': '2023-09-12',
    'so': "2021-11-22"
}


class MetaData:
    """
    A representation of the MetaData element of the GA4GH Phenopacket Schema

    :param created_by: identifier (such as ORCID id) of the person who created this Phenopacket
    :type created_by: str
    :param citation: PubMed identifier of the article from which the data for the phenopacket was taken, optional
    :type citation: Citation
    """

    def __init__(self, created_by, citation:Citation=None) -> None:
        """
        Constructor
        """
        self._created_by = created_by
        self._schema_version = "2.0"
        self._extref = None
        if citation is not None:
            self.set_external_reference(pmid=citation.pmid, pubmed_title=citation.title)
        self._citation = citation
        self._resource_d = defaultdict(Resource)

    def default_versions_with_hpo(self, version):
        """
        Add resources for HPO (with specified version), GENO, HGNC, and OMIM (with default versions)
        The HPO version can be easily obtained from the HpoParser using the get_version() function

        :param version: version of the Human Phenotype Ontology (HPO) used to create this phenopacket
        :type version: str
        :param pmid: PubMed identifier of the article from which the data for the phenopacket was taken, optional
        :type pmid: str
        :param pubmed_title: title of the article (if any), for use in the Resource section
        :type pubmed_title: str
        """
        self.geno()
        self.hgnc()
        self.omim()
        self.sequence_ontology()
        self.hpo(version=version)

    def hpo(self, version):
        """
        :param version: version of the Human Phenotype Ontology (HPO) used to create this phenopacket
        :type version: str
        """
        self._resource_d["hp"] = Resource(resource_id="hp",
                                        name="human phenotype ontology",
                                        namespace_prefix="HP",
                                        iriprefix="http://purl.obolibrary.org/obo/HP_",
                                        url="http://purl.obolibrary.org/obo/hp.owl",
                                        version=version)

    def geno(self, version=default_versions.get('geno')):
        """_summary_
        GENO is used for three terms: homozygous, heterozygous, hemizygous
        For this reason, we use a default version, since we assume these terms will not change.
        Args:
            version (str, optional): GENO version. Defaults to "2022-03-05".

        Returns:
            _type_: Resource object representing GENO
        """
        self._resource_d["geno"] = Resource(resource_id="geno",
                                            name="Genotype Ontology",
                                            namespace_prefix="GENO",
                                            iriprefix="http://purl.obolibrary.org/obo/GENO_",
                                            url="http://purl.obolibrary.org/obo/geno.owl",
                                            version=version)

    def hgnc(self, version=default_versions.get('hgnc')):
        self._resource_d["hgnc"] = Resource(resource_id="hgnc",
                                            name="HUGO Gene Nomenclature Committee",
                                            namespace_prefix="HGNC",
                                            iriprefix="https://www.genenames.org/data/gene-symbol-report/#!/hgnc_id/",
                                            url="https://www.genenames.org",
                                            version=version)

    def omim(self, version=default_versions.get('omim')):
        self._resource_d["omim"] = Resource(resource_id="omim",
                                            name="An Online Catalog of Human Genes and Genetic Disorders",
                                            namespace_prefix="OMIM",
                                            iriprefix="https://www.omim.org/entry/",
                                            url="https://www.omim.org",
                                            version=version)

    def mondo(self, version=default_versions.get('mondo')):
        """
        Add a resource for Mondo to the current MetaData object
        :param version: the Mondo version
        """
        self._resource_d["mondo"] = Resource(resource_id="mondo",
                                            name="Mondo Disease Ontology",
                                            namespace_prefix="MONDO",
                                            iriprefix="http://purl.obolibrary.org/obo/MONDO_",
                                            url="http://purl.obolibrary.org/obo/mondo.obo",
                                            version=version)

    def sequence_ontology(self, version=default_versions.get("so")):
        self._resource_d["so"] = Resource(resource_id="so",
                                            name="Sequence types and features ontology",
                                            namespace_prefix="SO",
                                            iriprefix="http://purl.obolibrary.org/obo/SO_",
                                            url="http://purl.obolibrary.org/obo/so.obo",
                                            version=version)

    def add_reference(self, resource:Resource) -> None:
        resource_id = resource.id
        self._resource_d[resource_id] = resource

    def set_external_reference(self, pmid, pubmed_title) -> None:
        """
        Set the external reference for this phenopacket/individual to be the PubMed identifier and title of an article

        :param pmid: The PubMed identifier of the publication from which the data was derived
        :type pmid: str
        :param pubmed_title: The title of the publication
        :type pubmed_title: str
        """
        self._extref = PPKt.ExternalReference()
        self._extref.id = pmid
        pm = pmid.replace("PMID:", "")
        self._extref.reference = f"https://pubmed.ncbi.nlm.nih.gov/{pm}"
        self._extref.description = pubmed_title

    def get_citation(self):
        return self._citation

    def get_pmid(self)->str:
        """
        :returns: The PubMed identifier
        :rtype: str:
        :raises ValueError: Throw an error if no PMID is available
        """
        if self._extref is not None:
            if self._extref.id.startswith("PMID"):
                return self._extref.id
            else:
                raise ValueError(f"Malformed PMID in external reference: {self._extref.id}")
        else:
            raise ValueError("Could not get PMID because MetaData._extref was None")

    def get_created_by(self) -> str:
        return self._created_by


    def to_ga4gh(self):
        """
        Use a time stamp for the current instant
        :returns: A MetaData formated as a GA4GH Phenopacket Schema message
        :rtype: PPkt.MetaData
        """
        metadata = PPKt.MetaData()
        metadata.created_by = self._created_by
        now = time.time()
        seconds = int(now)
        nanos = int((now - seconds) * 10 ** 9)
        timestamp = protobuf.timestamp_pb2.Timestamp(seconds=seconds, nanos=nanos)
        metadata.created.CopyFrom(timestamp)
        metadata.phenopacket_schema_version = self._schema_version
        for _, resource in self._resource_d.items():
            res = PPKt.Resource()
            res.id = resource.id
            res.name = resource.name
            res.namespace_prefix = resource.namespace_prefix
            res.iri_prefix = resource.iri_prefix
            res.url = resource.url
            res.version = resource.version
            metadata.resources.append(res)
        if self._extref is not None:
            metadata.external_references.append(self._extref)
        return metadata



from .hp_term import HpTerm
from .column_mapper import ColumnMapper
from .thresholder import Thresholder
from typing import List
import pandas as pd
import pkg_resources
from collections import defaultdict
import math


class ThresholdedColumnMapper(ColumnMapper):

    def __init__(self, column_name, thresholder:Thresholder):
        super().__init__(column_name=column_name)
        self._thresholder = thresholder

    def map_cell(self, cell_contents) -> List[HpTerm]:
        hpterm = self._thresholder.map_value(cell_contents=cell_contents)
        if hpterm is not None:
            return [hpterm]
        else:
            return []


    def preview_column(self, df:pd.DataFrame):
        if not isinstance(df, pd.DataFrame):
            raise ValueError("df argument must be pandas DataFrame, but was {type(column)}")
        column = df[self._column_name]
        mapping_counter = defaultdict(int)
        for _, value in column.items():
            results = self.map_cell(str(value))
            if len(results) > 0:
                hpterm = results[0]
                mapped = f"{hpterm.hpo_term_and_id}: {hpterm.display_value}"
                mapping_counter[mapped] += 1
        dlist = []
        colname = f"mapping: {self._thresholder.get_reference_range()}"
        for k, v in mapping_counter.items():
            d = {colname: k, "count": str(v)}
            dlist.append(d)
        return pd.DataFrame(dlist)




from .hp_term import HpTerm
from .column_mapper import ColumnMapper
from .hpo_cr import HpoConceptRecognizer
from typing import List, Set
import pandas as pd
import re
from collections import defaultdict


class OptionColumnMapper(ColumnMapper):
    """Class to map the contents of a table cell to one or more options (HPO terms)

    This mapper should be used if the column has a set of multiple defined items (strings) representing HPO terms.
    The excluded_d argument should be used if the column includes excluded (negated) HPO terms
    :param column_name: name of the column in the pandas DataFrame
    :type column_name: str
    :param concept_recognizer: HpoConceptRecognizer for text mining
    :type  concept_recognizer: pyphetools.creation.HpoConceptRecognizer
    :param option_d: dictionary with key: string corresponding to original table, value: corresponding HPO term label
    :type option_d:TypedDict[str,str]
    :param excluded_d: dictionary with key: similar to option_d but for excluded HPO terms, optional
    :type excluded_d:TypedDict[str,str]

    :param omitSet: set of strings to be excluded from concept recognition
    :type omitSet: Set[str]
    """

    def __init__(self, column_name:str, concept_recognizer, option_d, excluded_d=None, omitSet:Set[str]=None):
        """Constructor
        """
        super().__init__(column_name=column_name)
        # Either have self._option_d be an empty dictionary or it must be a valid dictionary
        if option_d is None or not isinstance(option_d, dict):
            raise ValueError(f"option_d argument must be dictionary but was {type(option_d)}")
        self._option_d = option_d
        if not isinstance(concept_recognizer, HpoConceptRecognizer):
            raise ValueError("concept_recognizer arg must be HpoConceptRecognizer but was {type(concept_recognizer)}")
        self._hpo_cr = concept_recognizer
        if excluded_d is None:
            excluded_d = defaultdict()
        self._excluded_d = excluded_d
        if omitSet is None:
            omitSet = set()
        elif not isinstance(omitSet, set):
            raise ValueError(f"argument 'omitSet' must be a Python set but was {type(omitSet)}")
        self._omit_set = omitSet


    def map_cell(self, cell_contents) -> List[HpTerm]:
        """parse a single table cell

        :param cell_contents: contents of a cell of the original file
        :type cell_contents: str
        :returns: list of HPO matches
        :rtype: List[HpTerm]
        """
        for excl in self._omit_set:
            cell_contents = cell_contents.replace(excl, " ")
        # results is a list of HpTerm objects
        results = []
        if self._excluded_d is not None and len(self._excluded_d) > 0:
            excluded_results = self._hpo_cr.parse_cell_for_exact_matches(cell_text=cell_contents, custom_d=self._excluded_d)
            if excluded_results is not None and len(excluded_results) > 0:
                for er in excluded_results:
                    er.excluded()
                    results.append(er)
                    # remove excluded terms from contents
                    cell_contents = cell_contents.replace(er.label.lower(), " ")
        results_obs = self._hpo_cr.parse_cell(cell_contents=cell_contents, custom_d=self._option_d)
        results.extend(results_obs)
        return results

    def preview_column(self, df:pd.DataFrame) -> pd.DataFrame:
        """
        Generate a pandas dataframe with a summary of parsing of the entire column

        This method is intended for use in developing the code for ETL of an input column.
        It is only needed for development and debugging.

        :param column: A single column from the input table
        :type column: pd.Series
        :returns: a pandas dataframe with one row for each entry of the input column
        :rtype: pd.DataFrame
        """
        if not isinstance(df, pd.DataFrame):
            raise ValueError("df argument must be pandas DataFrame, but was {type(column)}")
        column = df[self._column_name]
        mapping_counter = defaultdict(int)
        for _, value in column.items():
            cell_contents = str(value)
            term_list = self.map_cell(cell_contents)
            for hpterm in term_list:
                mapped = f"{hpterm.hpo_term_and_id} ({hpterm.display_value})"
                mapping_counter[mapped] += 1
        dlist = []
        for k, v in mapping_counter.items():
            d = {"mapping": k, "count": str(v)}
            dlist.append(d)
        return pd.DataFrame(dlist)

    @staticmethod
    def autoformat(df: pd.DataFrame, hpo_cr:HpoConceptRecognizer, delimiter:str=",", omit_columns:List[str]=None) -> str:
        """Autoformat code from the columns so that we can easily copy-paste and change it.

        This method intends to save time by preformatting code the create OptionMappers. The following commands
        will print out skeleton Python code that can be easily adapted to create a mapper.

            result = OptionColumnMapper.autoformat(df=dft, concept_recognizer=hpo_cr, delimiter=",")
            print(result)

        :param df: data frame with the data about the individuals
        :type df: pd.DataFrame
        :param concept_recognizer: HpoConceptRecognizer for text mining
        :type  concept_recognizer: pyphetools.creation.HpoConceptRecognizer
        :param delimiter: the string used to delimit individual items in a cell (default: comma)
        :type delimiter: str
        :param omit_columns: names of columns to omit from this search
        :type omit_columns: List[str]
        :param df_name: name of the variable that corresponds to the dataframe
        :type df_name: str
        :returns: a string that should be displayed using a print() command in the notebook - has info about automatically mapped columns
        :rtype: str
        """
        lines = []
        if not isinstance(df, pd.DataFrame):
            raise ValueError(f"argument \"df\" must be a pandas DataFrame but was {type(df)}")
        if not isinstance(hpo_cr, HpoConceptRecognizer):
            raise ValueError(f"concept_recognizer arg must be HpoConceptRecognizer but was {type(hpo_cr)}")
        if omit_columns is None:
            omit_columns = set()
        elif isinstance(omit_columns, list):
            omit_columns = set(omit_columns)
        elif not isinstance(omit_columns, set):
            raise ValueError(f"If passed, omit_columns argument must be set or list but was {type(omit_columns)}")
        # df.shape[1] gives us the number of columns
        for y in range(df.shape[1]):
            original_column_name = str(df.columns[y])
            clean_column_name = str(df.columns[y]).lower().replace(", ","_").replace(' ', '_').replace("/", "_")
            col_name = clean_column_name.lower()
            if original_column_name in omit_columns:
                continue
            temp_dict = {}
            for i in range(len(df)):
                if len(str(df.iloc[i, y])) > 1:
                    for entry in str(df.iloc[i, y]).split(delimiter):
                        hpo_term = hpo_cr.parse_cell(entry.strip())
                        if len(hpo_term) > 0:
                            temp_dict[entry.strip()] = hpo_term[0].label
                        else:
                            temp_dict[entry.strip()] = 'PLACEHOLDER'

            # skip columns that are unlikely to be interesting for the OptionColumnMapper
            if "patient" in col_name:
                continue
            if "family" in col_name:
                continue
            if "gender" in col_name:
                continue
            if "mutation" in col_name or "varia" in col_name:
                continue
            if "age" == col_name or "sex" == col_name or "gender" == col_name:
                continue
            items_d_name = f"{col_name}_d"
            items_d_string = str(temp_dict).replace(',', ',\n')
            lines.append(f"{items_d_name} = {items_d_string}")
            lines.append("excluded = {}")
            lines.append(f"{col_name}Mapper = OptionColumnMapper(column_name=\"{original_column_name}\", concept_recognizer=hpo_cr, option_d={items_d_name}, excluded_d=excluded)")
            lines.append(f"column_mapper_list.append({col_name}Mapper)")
            lines.append(f"{col_name}Mapper.preview_column(df)")

            lines.append("")
        return "\n".join(lines)




from typing import List
from collections import defaultdict
import hpotk
from ..validation.validation_result import ValidationResult
from ..validation.validated_individual import ValidatedIndividual
from ..validation.cohort_validator import CohortValidator

from .html_table_generator import HtmlTableGenerator




class QcElement:
    """Helper class to provide additional context and order for displaying summaries of Qc events

    :param level: ERROR, WARNING, or INFORMATION
    :type level: str
    :param category: from the ValidationResult categories
    :type category: str
    """
    def __init__(self, level, category) -> None:
        self._level = level
        self._category = category

    @property
    def level(self):
        return self._level

    @property
    def category(self):
        return self._category

    @staticmethod
    def get_qc_element_list():
        elements = []
        elements.append(QcElement(level="ERROR", category="CONFLICT"))
        elements.append(QcElement(level="ERROR", category="INSUFFICIENT_HPOS"))
        elements.append(QcElement(level="ERROR", category="INCORRECT_ALLELE_COUNT"))
        elements.append(QcElement(level="ERROR", category="INCORRECT_VARIANT_COUNT"))
        elements.append(QcElement(level="ERROR", category="MALFORMED_ID"))
        elements.append(QcElement(level="ERROR", category="MALFORMED_LABEL"))
        elements.append(QcElement(level="ERROR", category="OBSERVED_AND_EXCLUDED"))
        elements.append(QcElement(level="WARNING", category="REDUNDANT"))
        elements.append(QcElement(level="INFORMATION", category="NOT_MEASURED"))
        return elements



class QcVisualizer:
    """Class degigned to create HTML summaries of the validation/QC of a cohort
    :param ontology: HPO ontology
    :type ontology: hpotk.MinimalOntology
    :param cohort_validator: Validator object that checks all individuals in a cohort
    :type cohort_validator: CohortValidator
    """
    def __init__(self, cohort_validator:CohortValidator) -> None:
        self._ontology = cohort_validator.get_ontology()
        if not isinstance(cohort_validator, CohortValidator):
            raise ValueError(f"cohort_validator argument must be CohortValidator object but was {type(cohort_validator)}")
        self._cohort_validator = cohort_validator


    def to_html(self) -> str:
        validated_individual_list = self._cohort_validator.get_validated_individual_list()
        html_lines = []
        n_individuals = len(validated_individual_list)
        n_individuals_with_errors = sum([1 for i in validated_individual_list if i.has_error()])
        html_lines.append("<h2>Cohort validation</h2>")
        if n_individuals_with_errors == 0:
            html_lines.append(f"<p>No errors found for the cohort with {n_individuals} individuals</p>")
        else:
            para = f"<p>Errors found with {n_individuals_with_errors} of {n_individuals} phenopackets.</p>"
            html_lines.append(para)
            errors = []
            for vi in validated_individual_list:
                if vi.has_error():
                    errors.extend(vi.get_validation_errors())
            errors = sorted(errors, key=lambda x : (x._error_level, x._category, x._message))
            header_fields = ValidationResult.get_header_fields()
            rows = [row.get_items_as_array() for row in errors]
            generator = HtmlTableGenerator(header_items=header_fields, rows=rows, caption="Error analysis")
            html_lines.append(generator.get_html())
        return "\n".join(html_lines)

    def to_summary_html(self) -> str:
        validated_individual_list = self._get_validated_individuals()
        qc_element_list = QcElement.get_qc_element_list()
        error_count = defaultdict(int)
        total_validation_issues = 0
        distinct_item_d = defaultdict(set)
        for vi in validated_individual_list:
            verrors = vi.get_validation_errors()
            total_validation_issues += vi.n_errors()
            for ve in verrors:
                error_count[ve.category] += 1
                distinct_item_d[ve.category].add(ve.message)
        html_lines = []
        n_individuals = len(validated_individual_list)
        n_individuals_with_errors = sum([1 for i in validated_individual_list if i.has_error()])
        html_lines.append("<h2>Cohort validation</h2>")
        if n_individuals_with_errors == 0:
            html_lines.append(f"<p>No errors found for the cohort with {n_individuals} individuals</p>")
        else:
            para = f"<p>Errors found with {n_individuals_with_errors} of {n_individuals} phenopackets.</p>"
            html_lines.append(para)

            header_fields = ["Level", "Error category", "Count"]
            rows = []
            for elem in qc_element_list:
                if elem.category in error_count:
                    count = error_count.get(elem.category)
                    row = [elem.level, elem.category, str(count)]
                    rows.append(row)
            generator = HtmlTableGenerator(header_items=header_fields, rows=rows, caption="Error counts")
            html_lines.append(generator.get_html())
            n_removed = self._cohort_validator.n_removed_individuals()
            n_error_free_i = self._cohort_validator.n_error_free_individuals()
            if n_removed == 0:
                para = f"<p>A total of {total_validation_issues} issues were fixed and no individual was removed from the cohort.</p>"
                html_lines.append(para)
            else:
                para = f"<p>A total of {total_validation_issues} issues were fixed and {n_removed} individuals were removed from the cohort because of irreparable errors. The cohort validator will return {n_error_free_i} individual objects without errors.</p>"
                html_lines.append(para)
                if "MALFORMED_LABEL" in distinct_item_d:
                    malformed_label_set = distinct_item_d.get("MALFORMED_LABEL")
                    if len(malformed_label_set) > 0:
                        malformed = "; ".join(malformed_label_set)
                        para = f"<p>The following malformed labels were found: {malformed}. These need to be corrected before continuing.</p>"
                        html_lines.append(para)
                if "REDUNDANT" in distinct_item_d:
                    redundant_label_set = distinct_item_d.get("REDUNDANT")
                    redundant = "; ".join(redundant_label_set)
                    para = f"<p>The following redundant terms were found: {redundant}. Redundant terms will be removed, keeping only one instance of the most specific term.</p>"
                    html_lines.append(para)
                if "CONFLICT" in distinct_item_d:
                    conflict_label_set = distinct_item_d.get("CONFLICT")
                    conflict = "; ".join(conflict_label_set)
                    para = f"<p>The following excluded terms were found to have a conflict with an observed descendent term: {conflict}. The ancestor terms will be removed.</p>"
                    html_lines.append(para)
                if "OBSERVED_AND_EXCLUDED" in distinct_item_d:
                    o_and_e_set = distinct_item_d.get("OBSERVED_AND_EXCLUDED")
                    o_and_e = "; ".join(o_and_e_set)
                    para = f"<p>The following terms were annotated as being both observed and excluded: {o_and_e}. This needs to be fixed manually.</p>"
                    html_lines.append(para)
                if "DUPLICATE" in distinct_item_d:
                    dup_set = distinct_item_d.get("DUPLICATE")
                    dup = "; ".join(dup_set)
                    para = f"<p>The following terms were annotated as duplicates: {dup}. This will be fixed automatically.</p>"
                    html_lines.append(para)
                html_lines.append(self._get_unfixable_error_table())
        return "\n".join(html_lines)


    def _get_unfixable_error_table(self):
        v_individuals_with_unfixable = self._cohort_validator.get_validated_individuals_with_unfixable_errors()
        html_lines = []
        html_lines.append("<h2>Individuals with unfixable errors</h2>")
        errors = []
        for vi in v_individuals_with_unfixable:
            if vi.has_unfixed_error():
                errors.extend(vi.get_unfixed_errors())
        errors = sorted(errors, key=lambda x : (x._error_level, x._category, x._message))
        header_fields = ValidationResult.get_header_fields()
        rows = [row.get_items_as_array() for row in errors]
        generator = HtmlTableGenerator(header_items=header_fields, rows=rows, caption="Error analysis")
        html_lines.append(generator.get_html())
        return "\n".join(html_lines)


    def _get_validated_individuals(self):
        validated_individual_list = self._cohort_validator.get_validated_individual_list()
        if not isinstance(validated_individual_list, list):
            raise ValueError(f"validated_individual_list argument must be a list but was {type(validated_individual_list)}")
        if len(validated_individual_list) == 0:
            raise ValueError(f"validated_individual_list argument was empty")
        if not isinstance(validated_individual_list[0], ValidatedIndividual):
            raise ValueError(f"validated_individual_list argument must be a list of ValidatedIndividual objects but was {type(validated_individual_list[0])}")
        return validated_individual_list






import pandas as pd
from typing import List, Dict
from .variant import Variant

ACCEPTABLE_GENOTYPES = {"heterozygous", "homozygous", "hemizygous"}



class VariantColumnMapper:
    """Column mapper for the variants identified in individuals

    To use this mapper, first create a dictionary with all variants in the cohort using the
    VariantValidator (for HGVS) and the StructuralVariant classes. The key to the variant is
    the cell contents of the column with variant information in the original table. The Values
    are the Variant objects (implemented as HgvsVariant or StructuralVariant). This mapper will
    split cells that contain multiple variants if needed and also will add genotype information.

    :param variant_d: Dictionary with all variants found in the column
    :type variant_d: Dict[str,Variant]
    :param variant_column_name: name of the variant column in the original table
    :type variant_column_name: str
    :param genotype_column_name: name of the genotype column in the original table, optional
    :type genotype_column_name: str
    :param delimiter: symbol used to separate variants if more than one variant is provided, optional
    :type delimiter: str
    """
    def __init__(self,
                variant_d,
                variant_column_name,
                genotype_column_name=None,
                default_genotype=None,
                delimiter=None) -> None:

        if default_genotype is not None and default_genotype not in ACCEPTABLE_GENOTYPES:
            raise ValueError(f"Did not recognize default genotype {default_genotype}")
        if not isinstance(variant_d, dict):
            raise ValueError(f"Argument variant_d must be a dictionary but was {type(variant_d)}")
        self._variant_d = variant_d
        self._default_genotype = default_genotype
        self._variant_column_name = variant_column_name
        self._genotype_column_name = genotype_column_name
        self._delimiter = delimiter


    def _get_genotype(self, genotype_contents):
        """
        Get a genotype string

        get the genotype from the argument (which comes from the genotype column if available) or from the
        default genotype.
        :returns: one of "heterozygous", "homozygous", "hemizygous", None
        """
        if genotype_contents is not None and genotype_contents.lower() in ACCEPTABLE_GENOTYPES:
            return genotype_contents.lower()
        else:
            return self._default_genotype


    def map_cell(self, cell_contents, genotype_contents=None, delimiter=None) -> List[Variant]:
        """
        Map the contents of a variant cell (and optionally a genotype cell).

        If the delimiter is not None, search for the delimiter and split the cell into two variants
        :param cell_contents: contents of the original table cell representing the variant string
        :type cell_contents: str
        :param genotype_contents: contents of the original table cell representing the genotype (allelic status), optional
        :type genotype_contents: str
        :param delimiter: string or character that splits the cell contents into multiple entries, e.g., ";", optional
        :type delimiter: str
        """
        if delimiter is None:
            delimiter = self._delimiter
        if delimiter is not None:
            items = [x.strip() for x in cell_contents.split(delimiter)]
        else:
            items = [cell_contents]
        variant_interpretation_list = []
        for item in items:
            if item in self._variant_d:
                variant: Variant = self._variant_d.get(item)
                gt = self._get_genotype(genotype_contents)
                if gt is not None:
                    variant.set_genotype(gt)
                interpretation = variant.to_ga4gh_variant_interpretation()
                variant_interpretation_list.append(interpretation)
            else:
                raise ValueError(f"Did not recognize variant string \"{item}\"")

        return variant_interpretation_list



    def preview_column(self, column) -> pd.DataFrame:
        if not isinstance(column, pd.Series):
            raise ValueError("column argument must be pandas Series, but was {type(column)}")
        dlist = []
        for _, value in column.items():
            variant_interpretation_list = self.map_cell(str(value))
            if len(variant_interpretation_list) == 0:
                dlist.append({"variant": "n/a"})
            else:
                result_strings = []
                for v in variant_interpretation_list:
                    result_strings.append(v.__str__())
                dlist.append({"variant": ": ".join(result_strings)})
        return pd.DataFrame(dlist)

    def get_variant_column_name(self):
        return self._variant_column_name

    def get_genotype_colname(self):
        return self._genotype_column_name





import os
import json
from collections import defaultdict
from typing import List, Dict

from google.protobuf.json_format import Parse

from ..creation.disease import Disease
from ..creation.hp_term import HpTerm
from ..creation.individual import Individual
from ..creation.metadata import MetaData
from .simple_age import SimpleAge

EMPTY_CELL = ""

import phenopackets as PPKt
from collections import defaultdict
import pandas as pd

class HpoaTableRow:
    def __init__(self, disease:Disease, hpo_term:HpTerm, publication, biocurator,  freq_num:int=None, freq_denom:int=None):
        self._disease_id = disease.id
        self._disease_label = disease.label
        self._phenotype_id = hpo_term.id
        self._phenotype_name = hpo_term.label
        self._publication = publication
        self._biocuration = biocurator
        if freq_num is not None and freq_denom is not None:
            self._frequency = f"{freq_num}/{freq_denom}"
        else:
            self._frequency = None

    def get_dict(self):
        if self._frequency is not None:
            frequency = self._frequency
        else:
            frequency = EMPTY_CELL
        d = {"#diseaseID": self._disease_id,
            "diseaseName": self._disease_label,
            "phenotypeID": self._phenotype_id,
            "phenotypeName": self._phenotype_name,
            "onsetID":EMPTY_CELL,
            "onsetName": EMPTY_CELL,
            "frequency": frequency,
            "sex": EMPTY_CELL,
            "negation" : EMPTY_CELL,
            "modifier": EMPTY_CELL,
            "description": EMPTY_CELL,
            "publication": self._publication,
            "evidence": "PCS",
            "biocuration": self._biocuration
            }
        return d


class OnsetTerm:
    def __init__(self, onset_term_id, onset_term_label, numerator, denominator):
        if not isinstance(numerator, int):
            raise ValueError(f"Malformed numerator (must be integer but was {numerator})")
        if not isinstance(denominator, int):
            raise ValueError(f"Malformed denominator (must be integer but was {denominator})")
        self._onset_term_id = onset_term_id
        self._onset_term_label = onset_term_label
        self._num = numerator
        self._denom = denominator

    @property
    def id(self):
        return self._onset_term_id

    @property
    def label(self):
        return self._onset_term_label

    def has_frequency(self):
        return self._num is not None and self._denom is not None

    @property
    def numerator(self):
        return self._num

    @property
    def denominator(self):
        return self._denom


class HpoaPmidCounter:
    def __init__(self) -> None:
        self._observed = defaultdict(int)
        self._measured = defaultdict(int)

    def increment_observed(self, hpo_id:str):
        self._observed[hpo_id] += 1

    def increment_measured(self, hpo_id:str):
        self._measured[hpo_id] += 1

    def get_observed_d(self):
        return self._observed

    def get_measured_d(self):
        return self._measured


class HpoaOnsetCounter:
    def __init__(self) -> None:
        self._onset = defaultdict(int)

    def increment_onset(self, hpo_id:str):
        self._onset[hpo_id] += 1

    def get_onset_numerator_d(self):
        return self._onset

    def get_onset_denominator(self):
        d = sum([v for v in self._onset.values()])
        if d == 0:
            raise ValueError(f"Attempt to use HpoaOnsetCounter with no data")
        return d



class HpoaTableCreator:
    """
    Create an HPO "small file" with the following fourteen columns
    1. #diseaseID
    2. diseaseName
    3. phenotypeID
    4. phenotypeName
    5. onsetID
    6. onsetName
    7. frequency
    8. sex
    9. negation
    10. modifier
    11. description
    12. publication
    13. evidence
    14. biocuration
    These should be tab separated fields.

    """
    def __init__(self, phenopacket_list, onset_term_d,  moi_d) -> None:
        self._phenopackets = phenopacket_list
        self._all_hpo_d = self._get_all_hpos()
        self._disease = self._get_disease() # only allow one disease, therefore this is a scalar value (string)
        self._hpo_counter_d = self._count_hpos()
        self._biocurator_d = self._get_biocurator_d()
        self._onset_rows = self._add_age_of_onset_terms(onset_term_d)
        self._moi_rows = self._add_moi_rows(moi_d)

    def _get_all_hpos(self) -> Dict[str,HpTerm]:
        """Get a dictionary of HpTerms, with key being HPO id and the value the corresponding HpTerm

        We use this to retrieve the label
        :returns: a dictionary, with key=HPO id, value: HpTerm
        :rtype: Dict[str, HpTerm]
        """
        all_hpo_d = {}
        for ppkt in self._phenopackets:
            for pf in ppkt.phenotypic_features:
                hpterm = HpTerm(hpo_id=pf.type.id, label=pf.type.label)
                all_hpo_d[hpterm.id] = hpterm
        print(f"We found a total of {len(all_hpo_d)} unique HPO terms")
        return all_hpo_d


    @staticmethod
    def get_pmid(ppkt):
        mdata = ppkt.meta_data
        if mdata.external_references is None:
            raise ValueError("MetaData must have external_references element for HPOA conversion")
        eref_list = mdata.external_references
        if len(eref_list) != 1:
            raise ValueError(f"MetaData must have exactly one external_references element for HPOA conversion but had {len(eref_list)}")
        eref = eref_list[0]
        pmid = eref.id
        if not pmid.startswith("PMID:"):
            raise ValueError(f"Malformed PMID: \"{pmid}\"")
        return pmid


    def _get_disease(self):
        disease_set = set()
        for ppkt in self._phenopackets:
            interpretations = ppkt.interpretations
            if len(interpretations) != 1:
                raise ValueError(f"Error: must have only a single disease for HPOA conversion but we found {len(interpretations)}")
            interpretation = interpretations[0]
            if interpretation.diagnosis is None:
                raise ValueError(f"Could not get diagnosis object from interpretation with id {interpretation.id}")
            diagnosis = interpretation.diagnosis
            disease = Disease(disease_id=diagnosis.disease.id, disease_label=diagnosis.disease.label)
            disease_set.add(disease)
        if len(disease_set) == 0:
            raise ValueError("Could not retrieve Disease for cohort")
        elif len(disease_set) > 1:
            raise ValueError(f"Error: must have only a single Disease for HPOA conversion but we found {len(disease_set)}")
        [disease] = disease_set
        print(f"Extracted disease: {disease}")
        return disease

    def _get_biocurator_d(self):
        """The unspoken assumption of this function is that there is just one biocurator per PMID.
        This will be true for phenopackets created by pyphetools.

        :returns: dictionary with key=PMID, value=biocurator
        :rtype: Dict[str,str]
        """
        biocurator_d = defaultdict()
        for ppkt in self._phenopackets:
            pmid = HpoaTableCreator.get_pmid(ppkt=ppkt)
            mdata = ppkt.meta_data
            created_by = mdata.created_by
            biocurator_d[pmid] = created_by
        return biocurator_d

    def _count_hpos(self):
        hpo_counter_d = defaultdict(HpoaPmidCounter)
        for ppkt in self._phenopackets:
            pmid = HpoaTableCreator.get_pmid(ppkt=ppkt)
            hpo_counter = hpo_counter_d.get(pmid)
            # if not yet present,  initialize with zero counts
            if  hpo_counter is None:
                hpo_counter = HpoaPmidCounter()
                hpo_counter_d[pmid] = hpo_counter
            for pf in ppkt.phenotypic_features:
                hpterm = HpTerm(hpo_id=pf.type.id, label=pf.type.label)
                hpo_counter.increment_measured(hpterm.id)
                if pf.excluded is not None and not pf.excluded:
                    hpo_counter.increment_observed(hpterm.id)
        return hpo_counter_d

    def _add_age_of_onset_terms(self, onset_term_d) -> List[HpoaTableRow]:
        onset_rows  = list() # reset
        for pmid, oterm_list in onset_term_d.items():
            # oterm = OnsetTerm(onset_term_id="HP:0003621", onset_term_label="Juvenile onset", numerator=num, denominator=denom)
            biocurator = self._biocurator_d.get(pmid)
            for oterm in oterm_list:
                hpo_onset_term = HpTerm(hpo_id=oterm.id, label=oterm.label)
                row = HpoaTableRow(disease=self._disease, hpo_term=hpo_onset_term, publication=pmid, biocurator=biocurator, freq_num=oterm.numerator, freq_denom=oterm.denominator)
                onset_rows.append(row)
        return onset_rows

    def _add_moi_rows(self, moi_d) -> List[HpoaTableRow]:
        """Add mode of inheritance information
        :param moi_d: dictionary with key: PMID, and value: List of MOI terms
        :type moi_d:Dict[str,List[str]]
        :returns: list of HPOA table rows
        :rtype: List[HpoaTableRow]
        """
        moi_rows = list()
        for pmid, hpterm_list in moi_d.items():
            biocurator = self._biocurator_d.get(pmid)
            for hpterm in hpterm_list:
                row = HpoaTableRow(disease=self._disease, hpo_term=hpterm, publication=pmid, biocurator=biocurator)
                moi_rows.append(row)
        return moi_rows


    def get_dataframe(self):
        rows = []
        column_names = ["#diseaseID", "diseaseName", "phenotypeID", "phenotypeName", "onsetID",
                        "onsetName", "frequency", "sex", "negation",  "modifier",
                        "description", "publication","evidence", "biocuration"]
        for pmid, counter in self._hpo_counter_d.items():
            biocurator = self._biocurator_d.get(pmid)
            measured_d = counter.get_measured_d()
            observed_d = counter.get_observed_d()
            # by construction, there can be no term in observed_d that is not in measured_d
            for hpo_id in measured_d:
                n = observed_d.get(hpo_id, 0)
                m = measured_d.get(hpo_id)
                hpo_term = self._all_hpo_d.get(hpo_id)
                row = HpoaTableRow(disease=self._disease, hpo_term=hpo_term, publication=pmid, biocurator=biocurator, freq_num=n, freq_denom=m)
                rows.append(row.get_dict())
        for onset_row in self._onset_rows:
            rows.append(onset_row.get_dict())
        for moi_row in self._moi_rows:
            rows.append(moi_row.get_dict())
        df = pd.DataFrame.from_records(data=rows, columns=column_names)
        return df

    def write_data_frame(self):
        df = self.get_dataframe()
        dlist = df["#diseaseID"].unique()
        if len(dlist) != 1:
            raise ValueError("Error - expected to get one disease but got {len(dlist)}")
        disease = dlist[0].replace(":", "-")
        fname = f"{disease}.tab"
        df.to_csv(fname, sep="\t", index=False)
        print(f"Wrote HPOA disease file to {fname}")


class HpoaTableBuilder:

    def __init__(self, indir=None, phenopacket_list=None) -> None:
        if indir is not None:
            if not os.path.isdir(indir):
                raise ValueError(f"indir argument {indir} must be directory!")
            self._indir = indir
            self._phenopackets = []
            for file in os.listdir(indir):
                fname = os.path.join(indir, file)
                if fname.endswith(".json") and os.path.isfile(fname):
                    with open(fname) as f:
                        data = f.read()
                        jsondata = json.loads(data)
                        ppack = Parse(json.dumps(jsondata), PPKt.Phenopacket())
                        self._phenopackets.append(ppack)
        elif phenopacket_list is not None:
            self._phenopackets = phenopacket_list
        else:
            raise ValueError("A valid value must be supplied for either \"indir\" or \"phenopacket_list\"")
        self._onset_term_d = defaultdict(list)
        self._moi_d = defaultdict(list)

    def embryonal_onset(self, pmid:str, num:int=None, denom:int=None):
        """Onset of disease at up to 8 weeks following fertilization (corresponding to 10 weeks of gestation).
        """
        oterm = OnsetTerm(onset_term_id="HP:0011460", onset_term_label="Embryonal onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self

    def fetal_onset(self, pmid:str, num:int=None, denom:int=None):
        """Onset prior to birth but after 8 weeks of embryonic development (corresponding to a gestational age of 10 weeks).
        """
        oterm = OnsetTerm(onset_term_id="HP:0011461", onset_term_label="Fetal onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self

    def second_trimester_onset(self, pmid:str, num:int=None, denom:int=None):
        """second trimester, which comprises the range of gestational ages from 14 0/7 weeks to 27 6/7 (inclusive)
        """
        oterm = OnsetTerm(onset_term_id="HP:0034198", onset_term_label="Second trimester onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self

    def late_first_trimester_onset(self, pmid:str, num:int=None, denom:int=None):
        """late first trimester during the early fetal period, which is defined as 11 0/7 to 13 6/7 weeks of gestation (inclusive).
        """
        oterm = OnsetTerm(onset_term_id="HP:0034199", onset_term_label="Late first trimester onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self

    def third_trimester_onset(self, pmid:str, num:int=None, denom:int=None):
        """third trimester, which is defined as 28 weeks and zero days (28+0) of gestation and beyond.
        """
        oterm = OnsetTerm(onset_term_id="HP:0034197", onset_term_label="Third trimester onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self

    def antenatal_onset(self, pmid:str, num:int=None, denom:int=None):
        """onset prior to birth
        """
        oterm = OnsetTerm(onset_term_id="HP:0030674", onset_term_label="Antenatal onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self

    def congenital_onset(self, pmid:str, num:int=None, denom:int=None):
        """A phenotypic abnormality that is present at birth.
        """
        oterm = OnsetTerm(onset_term_id="HP:0003577", onset_term_label="Congenital onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self

    def neonatal_onset(self, pmid:str, num:int=None, denom:int=None):
        """Onset of signs or symptoms of disease within the first 28 days of life.
        """
        oterm = OnsetTerm(onset_term_id="HP:0003623", onset_term_label="Neonatal onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self


    def infantile_onset(self, pmid:str, num:int=None, denom:int=None):
        """Onset of signs or symptoms of disease between 28 days to one year of life.
        """
        oterm = OnsetTerm(onset_term_id="HP:0003593", onset_term_label="Infantile onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self

    def childhood_onset(self, pmid:str, num:int=None, denom:int=None):
        """Onset of disease at the age of between 1 and 5 years.
        """
        oterm = OnsetTerm(onset_term_id="HP:0011463", onset_term_label="Childhood onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self

    def juvenile_onset(self, pmid:str, num:int=None, denom:int=None):
        """Onset of signs or symptoms of disease between the age of 5 and 15 years.
        """
        oterm = OnsetTerm(onset_term_id="HP:0003621", onset_term_label="Juvenile onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self

    def adult_onset(self, pmid:str, num:int=None, denom:int=None):
        """Onset of disease after 16 years  .
        """
        oterm = OnsetTerm(onset_term_id="HP:0003581", onset_term_label="Adult onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self

    def young_adult_onset(self, pmid:str, num:int=None, denom:int=None):
        """Onset of disease at the age of between 16 and 40 years.
        """
        oterm = OnsetTerm(onset_term_id="HP:0011462", onset_term_label="Young adult onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self

    def middle_age_onset(self, pmid:str, num:int=None, denom:int=None):
        """onset of symptoms at the age of 40 to 60 years.
        """
        oterm = OnsetTerm(onset_term_id="HP:0003596", onset_term_label="Middle age onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self

    def late_onset(self, pmid:str, num:int=None, denom:int=None):
        """Onset of symptoms after the age of 60 years.
        """
        oterm = OnsetTerm(onset_term_id="HP:0003584", onset_term_label="Late onset", numerator=num, denominator=denom)
        self._onset_term_d[pmid].append(oterm)
        return self

    def autosomal_recessive(self, pmid):
        moi_term = HpTerm(hpo_id="HP:0000007", label="Autosomal recessive inheritance")
        self._moi_d[pmid].append(moi_term)
        return self

    def autosomal_dominant(self, pmid):
        moi_term = HpTerm(hpo_id="HP:0000006", label="Autosomal dominant inheritance")
        self._moi_d[pmid].append(moi_term)
        return self

    def x_linked(self, pmid):
        moi_term = HpTerm(hpo_id="HP:0001417", label="X-linked inheritance")
        self._moi_d[pmid].append(moi_term)
        return self

    def x_linked_recessive(self, pmid):
        moi_term = HpTerm(hpo_id="HP:0001419", label="X-linked recessive inheritance")
        self._moi_d[pmid].append(moi_term)
        return self

    def x_linked_dominant(self, pmid):
        moi_term = HpTerm(hpo_id="HP:0001423", label="X-linked dominant inheritance")
        self._moi_d[pmid].append(moi_term)
        return self


    def build(self):
        return HpoaTableCreator(phenopacket_list=self._phenopackets, onset_term_d=self._onset_term_d, moi_d=self._moi_d)

    @staticmethod
    def from_individuals(individual_list:List[Individual], created_by:str):
        """Create builder object from list of individuals
        This can be easier than needed to create phenopacket objects in a workbook
        The method requires that each Individual object have a Citation

        :param individual_list: List of individuals to be summarized in HPOA format
        :type individual_list: List[Individual]
        :returns: HPOA table builder object
        :rtype: HpoaTableBuilder
        """
        ppkt_list = list()
        for i in individual_list:
            cite = i.get_citation()
            metadata = MetaData(created_by="ORCID:0000-0002-0736-9199", citation=cite)
            ppkt = i.to_ga4gh_phenopacket(metadata=metadata)
            ppkt_list.append(ppkt)
        return HpoaTableBuilder(phenopacket_list=ppkt_list)







import math
import os
import pandas as pd
from .hp_term import HpTerm
from .constants import Constants

current_dir = os.path.dirname(__file__)
THRESHOLDS_FILE = os.path.join(current_dir, "data", "thresholds.tsv");

class Thresholder:
    """A class to simplify the interpretation of thresholded values

    The class organizes the interpretation of numerical values for tests such as blood potassium. Results
    can be low, normal, or high, and there is an HPO term for each. The class also provides
    commonly used normal ranges. We note that many lab tests have different ranges for males and females or
    for adults and children. We do not attempt to model this, but instead provide ranges for adults.
    In cases where there are different ranges for males and females, we by default use the minimum and maximum
    value for each sex. It is possible to specify the range that should be used in the constructor method;
    in this case, the default values are overriden. In very complicated cases with multiple different
    normal ranges, consider using the OptionColumnMapper.
    """

    THRESHOLDER_MAP_NEEDS_INITIALIZATION = True

    THRESHOLDER_MAP = {}

    def __init__(self, unit:str, hpo_term_low=None, hpo_term_high=None, hpo_term_abn=None, threshold_low=None, threshold_high=None, threshold_low_male=None, threshold_high_male=None, threshold_low_female=None, threshold_high_female=None):
        """
        if hpo_term_low is not None and not isinstance(hpo_term_low, HpTerm):
            raise ValueError(f"hpo_term_low argument must be HpTerm but was {type(hpo_term_low)}")
        if hpo_term_high is not None and not isinstance(hpo_term_high, HpTerm):
            raise ValueError(f"hpo_term_high argument must be HpTerm but was {type(hpo_term_high)}")
        if hpo_term_abn is not None and not  not isinstance(hpo_term_abn, HpTerm):
            raise ValueError(f"hpo_term_abn argument must be HpTerm but was {type(hpo_term_abn)}")
        """
        self._hpo_low = hpo_term_low
        self._hpo_high = hpo_term_high
        self._hpo_term_abn = hpo_term_abn
        # The thresholds are allowed to be None but if they are given they must be numbers
        if threshold_low is not None and not isinstance(threshold_low, int) and not isinstance(threshold_low, float):
            raise ValueError(f"threshold_low argument must be integer or float but was {threshold_low}")
        if threshold_high is not None and not isinstance(threshold_high, int) and not isinstance(threshold_high, float):
            raise ValueError(f"threshold_high argument must be integer or float but was {threshold_high}")
        self._unit = unit
        self._threshold_low = self._set_threshold_to_float_or_none(threshold_low)
        self._threshold_high = self._set_threshold_to_float_or_none(threshold_high)

    def _set_threshold_to_float_or_none(self, thresh):
        """in the input file, a threshold value can be set to na or n/a. If so, we set the value to None.
        Otherwise, we convert all values to floats
        """
        if thresh is not None:
            return float(thresh)
        else:
            return None


    def set_unit(self, unit) -> None:
        """
        :param unit: a user-defined unit. This can be useful, for instance, if a paper uses mg/dL instead of SI units
        :type unit: str
        """
        self._unit = unit

    def set_high_threshold(self, value):
        """Set a user-defined high threshold value. This can be useful, for instance, if a paper uses mg/dL instead of SI units
        """
        value = float(value)
        self._threshold_high = value

    def set_low_threshold(self, value):
        """Set a user-defined low threshold value. This can be useful, for instance, if a paper uses mg/dL instead of SI units
        """
        value = float(value)
        self._threshold_low = value

    def get_reference_range(self):
        return f"{self._threshold_low}-{self._threshold_high} {self._unit}"


    def _value_is_high(self, value):
        if self._threshold_high is None:
            return False
        if self._hpo_high is None:
            return False
        return self._threshold_high < value

    def _value_is_low(self, value):
        if self._threshold_low is None:
            return False
        if self._hpo_low is None:
            return False
        return self._threshold_low > value

    def _value_is_normal(self, value):
        if self._threshold_high is None or self._threshold_low is None:
            return False
        if self._hpo_term_abn is None:
            return False
        return  value >= self._threshold_low and value <= self._threshold_high


    def _non_measured_term(self):
        if self._hpo_term_abn is not None:
            return HpTerm(hpo_id=self._hpo_term_abn.id, label=self._hpo_term_abn.label, measured=False)
        elif self._hpo_high is not None:
            return HpTerm(hpo_id=self._hpo_high.id, label=self._hpo_high.label, measured=False)
        elif self._hpo_low is not None:
            return HpTerm(hpo_id=self._hpo_low.id, label=self._hpo_low.label, measured=False)
        else:
            # should never happen
            raise ValueError("No HPO Term found for unmeasured")


    def map_value(self, cell_contents) -> HpTerm:
        if isinstance(cell_contents, str):
            contents = cell_contents.strip()
            if contents.lower() == "nan":
                return HpTerm(hpo_id=self._hpo_id, label=self._hpo_label, observed=False)
        elif isinstance(cell_contents, int):
            contents = cell_contents
        elif isinstance(cell_contents, float):
            if math.isnan(cell_contents):
                return HpTerm(hpo_id=self._hpo_id, label=self._hpo_label, observed=False)
            contents = cell_contents
        else:
            raise ValueError(
                f"Malformed cell contents for ThresholdedColumnMapper: {cell_contents}, type={type(cell_contents)}")
        try:
            value = float(contents)
            if self._value_is_high(value=value):
                return self._hpo_high
            elif self._value_is_low(value=value):
                return self._hpo_low
            elif self._value_is_normal(value=value):
                return HpTerm(hpo_id=self._hpo_term_abn.id, label=self._hpo_term_abn.label, observed=False)
            else:
                return self._non_measured_term()
        except Exception as exc:
            return self._non_measured_term()


    @staticmethod
    def _get_hpo_term_or_none(hpo_id, hpo_label):
        """It is OK if we do not have threshold terms for all three possible slots (Abnormal, High, Low)
        If we do not have a term, then we return None and the thresholder object will know what to do
        """
        # It is OK if there is a blank cell or a cell with na
        if hpo_id is None or hpo_label is None:
            return None
        if len(hpo_id) == 0 or len(hpo_label) == 0:
            return None
        if hpo_id == 'na' or hpo_label == 'na':
            return None
        # if the hpo_id is present, it must be well formed
        if not hpo_id.startswith('HP:') or len(hpo_id) != 10:
            raise ValueError(f"Malformed HP id: \"{hpo_id}\"")
        # we should be good to go when we get here
        return HpTerm(hpo_id=hpo_id, label=hpo_label)

    @staticmethod
    def _get_threshold_or_none(thresh):
        """the threshold cell can be empty or na -- this is OK, return None
        Otherwise return a float
        """
        if thresh is None:
            return None
        if isinstance(thresh, str):
            if len(thresh) == 0 or thresh == 'na' or thresh == "n/a":
                return None
        if isinstance(thresh, float) and math.isnan(thresh):
            return None
        return float(thresh)


    @staticmethod
    def _initialize_map():
        #current_dir = os.path.dirname(__file__)
        #thresholds_file = os.path.join(current_dir, "data", "thresholds.tsv");
        df = pd.read_csv(THRESHOLDS_FILE, delimiter="\t")
        for _, row in df.iterrows():
            label = row["label"]
            hpo_abn_label  = row["hpo_abn_label"]
            hpo_abn_id  = row["hpo_abn_id"]
            hpo_low_label  = row["hpo_low_label"]
            hpo_low_id  = row["hpo_low_id"]
            hpo_high_label  = row["hpo_high_label"]
            hpo_high_id   = row["hpo_high_id"]
            unit   = row["unit"]
            low  = row["low"]
            high  = row["high"]
            # Reference = row["Reference"] -- not needed here, for human consumption only
            hp_abnormal = Thresholder._get_hpo_term_or_none(hpo_id=hpo_abn_id, hpo_label=hpo_abn_label)
            hp_low = Thresholder._get_hpo_term_or_none(hpo_id=hpo_low_id, hpo_label=hpo_low_label)
            hp_high = Thresholder._get_hpo_term_or_none(hpo_id=hpo_high_id, hpo_label=hpo_high_label)
            low_thresh = Thresholder._get_threshold_or_none(low)
            high_thresh = Thresholder._get_threshold_or_none(high)
            thresh = Thresholder(unit=unit, hpo_term_abn=hp_abnormal, hpo_term_high=hp_high, hpo_term_low=hp_low, threshold_high=high_thresh, threshold_low=low_thresh)
            Thresholder.THRESHOLDER_MAP[label] = thresh
        Thresholder.THRESHOLDER_MAP_NEEDS_INITIALIZATION = False





    @staticmethod
    def _get_thresholder(thresholder_label, unit=None, low_thresh=None, high_thresh=None):
        if Thresholder.THRESHOLDER_MAP_NEEDS_INITIALIZATION:
            Thresholder._initialize_map()
        if not thresholder_label in Thresholder.THRESHOLDER_MAP:
            raise ValueError(f"Could not find thresholder object for \"{thresholder_label}\"")
        thresh = Thresholder.THRESHOLDER_MAP.get(thresholder_label)
        if unit is not None:
            thresh.set_unit(unit)
        if low_thresh is not None:
            thresh.set_low_threshold(low_thresh)
        if high_thresh is not None:
            thresh.set_high_threshold(high_thresh)
        return thresh


    #################
    ## Static methods for commonly used lab tests.
    ## data in data/thresholds.tsv

    @staticmethod
    def albumin_urine(unit=None, low_thresh=None, high_thresh=None):
        """Albuminuria - Urine Albumin-Creatinine Ratio (uACR)
        """
        return Thresholder._get_thresholder("albumin urine", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)

    @staticmethod
    def ALT_blood(unit=None, low_thresh=None, high_thresh=None):
        """ALT blood
        """
        return Thresholder._get_thresholder("ALT blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)

    @staticmethod
    def AST_blood(unit=None, low_thresh=None, high_thresh=None):
        """AST blood
        """
        return Thresholder._get_thresholder("AST blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)


    @staticmethod
    def alkaline_phophatase_blood(unit=None, low_thresh=None, high_thresh=None):
        """Alkaline phosphatase in the blood circulation (U/L)
        """
        return Thresholder._get_thresholder("alkaline phosphatase blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)

    @staticmethod
    def ALT_blood(unit=None, low_thresh=None, high_thresh=None):
        """ALT_blood in the blood circulation (U/L)‚
        """
        return Thresholder._get_thresholder("ALT blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)

    @staticmethod
    def AST_blood(unit=None, low_thresh=None, high_thresh=None):
        """AST_blood in the blood circulation (U/L)‚
        """
        return Thresholder._get_thresholder("AST blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)

    @staticmethod
    def calcium_blood(unit=None, low_thresh=None, high_thresh=None):
        """Calcium in the blood circulation 8.5 to 10.2 mg/dL (2.13 to 2.55 millimol/L).
        """
        return Thresholder._get_thresholder("calcium blood",unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)


    @staticmethod
    def creatine_kinase_blood(unit=None, low_thresh=None, high_thresh=None):
        """CK females is 30 to 145 U/L; males 55 to 170 U/L
        """
        return Thresholder._get_thresholder("creatine kinase blood", unit=unit, low_thresh=low_thresh, high_thresh=high_thresh)


    @staticmethod
    def creatinine_blood(unit=None, low_thresh=None, high_thresh=None):
        """Serum crea,  0.6-1.2 mg/dL in adult males and 0.5-1.1 mg/dL in adult females
        """
        return Thresholder._get_thresholder("creatinine blood", unit=unit, low_thresh=low_thresh, high_thresh=high_thresh)

    @staticmethod
    def free_fatty_acid_blood(unit=None, low_thresh=None, high_thresh=None):
        """free fatty acid
        """
        return Thresholder._get_thresholder("free fatty acid blood", unit=unit, low_thresh=low_thresh, high_thresh=high_thresh)

    @staticmethod
    def glucose_blood(unit=None, low_thresh=None, high_thresh=None):
        """fasting blood glucose; 70 mg/dL (3.9 mmol/L) and 100 mg/dL (5.6 mmol/L).
        """
        return Thresholder._get_thresholder("glucose blood", unit=unit, low_thresh=low_thresh, high_thresh=high_thresh)

    @staticmethod
    def hemoglobin_A1c(unit=None, low_thresh=None, high_thresh=None):
        """hemoglobin_A1c
        """
        return Thresholder._get_thresholder("hemoglobin_A1c", unit=unit, low_thresh=low_thresh, high_thresh=high_thresh)



    @staticmethod
    def HDL_cholesterol_blood(unit=None, low_thresh=None, high_thresh=None):
        """HDL cholesterol blood
        """
        return Thresholder._get_thresholder("HDL cholesterol blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)

    @staticmethod
    def insulin_blood(unit=None, low_thresh=None, high_thresh=None):
        """insulin blood
        """
        return Thresholder._get_thresholder("insulin blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)


    @staticmethod
    def lactate_blood(unit=None, low_thresh=None, high_thresh=None):
        """Serum lactate (0.5-1 mmol/L)
        """
        return Thresholder._get_thresholder("lactate blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)

    @staticmethod
    def LDL_cholesterol_blood(unit=None, low_thresh=None, high_thresh=None):
        """LDL cholesterol blood
        """
        return Thresholder._get_thresholder("LDL cholesterol blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)

    @staticmethod
    def NTproBNP_blood(unit=None, low_thresh=None, high_thresh=None):
        """NT-proBNP < 100 picograms per milliliter (pg/mL)
        """
        return Thresholder._get_thresholder("NT-proBNP blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)

    @staticmethod
    def sodium_blood(unit=None, low_thresh=None, high_thresh=None):
        """Sodium in the blood circulation 135-145 mEq/L
        """
        return Thresholder._get_thresholder("sodium blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)

    @staticmethod
    def potassium_blood(unit=None, low_thresh=None, high_thresh=None):
        """Serum potassium 3.5 to 5.2 mEq/L (adults)
        """
        return Thresholder._get_thresholder("potassium blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)

    @staticmethod
    def total_cholesterol_blood(unit=None, low_thresh=None, high_thresh=None):
        """total cholesterol blood
        """
        return Thresholder._get_thresholder("total cholesterol blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)


    @staticmethod
    def triglyceride_blood(unit=None, low_thresh=None, high_thresh=None):
        """triglyceride blood
        """
        return Thresholder._get_thresholder("triglyceride blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)

    @staticmethod
    def uric_acid_blood(unit=None, low_thresh=None, high_thresh=None):
        """uric acid blood
        """
        return Thresholder._get_thresholder("uric acid blood", unit=unit, low_thresh=low_thresh, high_thresh= high_thresh)




